# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.


class _HowtoClass:
    """The help class does nothing but offer docstring for
    function that have names related to potential activities.
    """
    def __init__(self):
        pass

    def start(self):
        """
        To start a simulation simply create a model object through:

        >>> model = emerge.Simulation('MyProjectName')

        Optionally, you can use a context manager for a more explicit handling of exiting the GMSH api and storing data after simulations.

        >>> with emerge.Simulation('MyProjectName') as model:

        """
        pass

    def make_geometry(self):
        """
        To make geometries in EMerge, you can use the various Geometry options in the .geo module.
        For example

        >>> box = emerge.geo.Box(...)
        >>> sphere = emerge.geo.Sphere(...)
        >>> pcb_layouter = emerge.geo.PCBLayout(...)
        >>> plate = emerge.geo.Plate(...)
        >>> cyl = emerge.geo.Cylinder(...)

        After making geometries, you should pass all of them to
        the simulation object
        >>> model.commit_geometry(box, sphere,...)

        You can also directly store geometries into the model class by treating
        it as a list (using __getitem__ and __setitem__)
        >>> model['box'] = emerge.geo.Box(...)

        In this case, all geometries will be automatically included. You shoul still call:
        >>> model.commit_geometry()

        """

        pass

    def select_face(self):
        """
        Faces can be selected in multiple ways. First, many native geometries have face definitions:
        >>> front_face = box.face('front')
        
        The naming convention is left(-X), right(+X), front(-Y), back(+Y), bottom(-Z), top(+Z)
        All outside faces can be selected using
        >>> outside = object.boundary()
        
        If objects are the results from operations, you can access the faces from the
        source objects using the optional tool argument
        >>> cutout = emerge.geo.subtract(sphere,box)
        >>> face = cutout.face('front', tool=box)
        
        Exclusions or specific isolations can be added with optional arguments.
        There is also a select object in your Simulation class that has various convenient selection options
        >>> faces = model.select.face.inlayer()
        >>> faces = model.select.inplane()
        >>> faces = model.select.face.near(x,y,z)
        
        """
        pass

    def boundary_conditions(self):
        """
        Boundary conditions can be added by first creating them and
        then passing them to the physics object. 
        >>> portbc = emerge.bc.ModalPort(...)
        >>> abc = emerge.bc.AbsorbingBoundary(...)
        >>> pec = emerge.bc.PEC(...)
        
        Then you should pass them to the physics object with comma separation
        and/or list comprehension (*bcs)
        >>> model.mw.assign(bc1, bc2, bc3, *bcs)

        """

    def run_sweep(self):
        """
        You can run a frequency domain study by simply calling:\
        
        >>> results = model.mw.run_sweep(...)

        You can distribute your frequency sweep across multiple threads using
        
        >>> results = model.mw.run_sweep(parallel=True, njobs=3)

        The frequency domain study will return an MWSimData object that contains all data.
        """
        pass

    def access_data(self):
        """
        Data from a frequency domain study is contained in the MWSimData object.
        Each simulation iteration is a separate MWDataSet object with all relevant parameters included.

        You can select an individual dataset based on the iteration number using:
        >>> dataset = results.item(3)
        
        This will select the 4th result. You can also select one by a specific value using
        >>> dataset = results(freq=3e9)
        
        The numbers must be exact. You can also approximately select a value using:
        >>> dataset = results.find(freq=3.9)
        
        Datasets contain various parameters such as the simulation frequency and associated k0:
        >>> k0 = dataset.k0
        >>> S11 = dataset.S(1,1)
        
        You can select all results along a specific axis using:
        >>> freq, S11 = results.ax('freq').S(1,1)
        
        If you ran a parameter sweep, this will select whatever value it finds.
        You can also select across multiple dimensions
        >>> freq, param, S11 = results.ax('freq','param').S(1,1)
        
        A dataset also contains all spatial data. You can probe the E-field or H-field using
        the interpolate method. This will compute the E and H fields and store them into the
        dataset after which they can be accessed
        >>> Ex, Ey, Ez = dataset(freq=1e9).interpolate(xs, ys, zs).E
        >>> Ex = dataset(freq=1e9).interpolate(xs, ys, zs).Ex
        
        You can automate the generation of sample coordinates and field values for
        specific plot instructions
        >>> X, Y, Z, Eyreal = dataset(freq=1e9).cutplane(ds=0.002, z=0.005).scalar('Ey','real')

        """
        pass

    def save_and_load(self):
        """
        You can save your project data by setting save_file to True:
        >>> model = emerge.Simulation(..., save_file=True)

        Whenever you want, you can save all data by calling the .save() method

        >>> model.save()

        If you run your simulation in a context manager, the data will be saved
        automatically when the context is done. By default, if your Python script crashes
        or ends, the data will also automatically be saved. Data is saved in a folder. 
        
        You can load the data from a simulation using:

        >>> model = emerge.Simulation(..., load_file=True)

        The data from a simulation can be found in:

        >>> results = model.data
        >>> results = model.mw.freq_data # the same


        """
        pass

    def parameter_sweep(self):
        """To run a parameter sweep, you can simply use the
        parameter_sweep iterator method of your model:
        
        >>>
        param1 = np.linspace(...)
        param2 = np.linspace(...)
        for p1, p2 in model.parameter_sweep(..., param_A=param1, param_B=param2):
            # run simulation
        
        The parameters will be automatically included in all simulation runs.
        The name will be the same as the keyword argument in the parameter sweep.
        You can extract the S-parameters for example using

        >>> freq, params, S11 = result.ax('freq','param_A').S(1,1)
        """
    
    def compute_farfield(self):
        """
        A farfield can be computed in any simulation but it only really
        represents something if the model has an absorbing boundary or PML.
        To compute the farfield on a single arc we will use the farfield_2d
        method of the MWDataSet class

        >>> theta, E, H = data.find(freq=1e9).farfield_2d(refdir, planedir, faces)

        The first argument should be the reference direction (angle = 0 degrees)
        The second argument is the plane normal. The arc is drawn as a semi-circle (or full)
        where the angle 0deg is in the reference direction.
        The faces argument should be a selection of faces to integrate the farfield over.
        
        Currently, EMerge only allows for Robin boundary conditions (ABC) on flat surfaces.

        """
        pass


# import shapely as shp
# import numpy as np


# def rectangle(p1: tuple, p2: tuple) -> shp.Polygon:
#     x1, y1 = p1
#     x2, y2 = p2
#     x1, x2 = sorted([x1, x2])
#     y1, y2 = sorted([y1, y2])
#     return shp.Polygon(((x1, y1), (x2, y1), (x2, y2), (x1, y2)))

# def rect_center(center: tuple, width: float, height: float):
#     x0, y0 = center
#     w, h = width, height
#     return shp.Polygon(((x0-w/2,y0-h/2),(x0+w/2,y0-h/2),(x0+w/2,y0+h/2),(x0-w/2,y0+h/2)))

# def polygon(points) -> shp.Polygon:
#     return shp.Polygon(points)

# def circle(center: tuple, radius: float, segments: int, ang_range: tuple = (0,360)):
#     x,y = center
#     points = []
#     path = []
#     angs = np.linspace(ang_range[0],ang_range[1],segments+1)

#     angs = angs*np.pi/180
#     for i in range(segments+1):
#         ang = angs[i]
#         points.append((x+radius*np.cos(ang),y+radius*np.sin(ang)))
#         path.append((x+radius*np.cos(ang),y+radius*np.sin(ang)))

#     if ang_range[0] != ang_range[1]%360:
#         points = [center,] + points
#     else:
#         points = points[:-1]
#     path = path + [path[0],]


#     return shp.Polygon(points), shp.LineString(path)

# def transform(polygon: shp.Polygon, transformation: callable, reverse=False) -> shp.Polygon:
#     ''' Transforms the provided shapely polygon by applying the transformation f(x,y) to each coordinate'''
#     ex, ey = polygon.exterior.xy
#     interiors = [interior.xy for interior in polygon.interiors]
#     new_poly = shp.Polygon(transformation(x,y) for x,y in zip(ex, ey))
#     for interior in interiors:
#         print(f'Type of interior = {type(interior)}, {interior}')
#         ix, iy = interior
#         new_poly = new_poly.difference(shp.Polygon(transformation(x,y) for x,y in zip(ix, iy)))
#     if reverse:
#         new_poly = new_poly.reverse()
#     return new_poly


# def move(polygon: shp.Polygon, dx: float, dy: float) -> shp.Polygon:
#     func = lambda x,y: (x+dx, y+dy)
#     return transform(polygon, func)

# def array(polygon: shp.Polygon, direction: tuple, N: int, include_original: bool = True) -> list[shp.Polygon]:
#     polys = []
#     for i in range(N+1):
#         polys.append(move(polygon, direction[0]*i, direction[1]*i))
#     if not include_original:
#         polys = polys[1:]
#     return polys

# def mirror(polygon: shp.Polygon, origin: tuple, axis: tuple) -> shp.Polygon:
#     ax, ay = axis[0]/np.sqrt(axis[0]**2 + axis[1]**2), axis[1]/np.sqrt(axis[0]**2 + axis[1]**2)
#     def _mir_transform(x, y):
#         dotprod = (x-origin[0])*ax + (y-origin[1])*ay
#         x2 = x - dotprod*ax*2
#         y2 = y - dotprod*ay*2
#         return x2, y2
#     return transform(polygon, _mir_transform, reverse=True)
    
# def rasterize(polygons: list[shp.Polygon], gridsize: float) -> list[shp.Polygon]:
#     output_polygons = []
#     grid = lambda x: np.round(np.array(x)/gridsize)*gridsize
#     for poly in polygons:
#         ex, ey = poly.exterior.xy
#         interiors = [interior.xy for interior in poly.interiors]
#         new_poly = shp.Polygon((x,y) for x,y in zip(grid(ex), grid(ey)))
#         for ix, iy in interiors:
#             new_poly.difference(shp.Polygon((x,y) for x,y in zip(grid(ix), grid(iy))))
#         output_polygons.append(new_poly)
#     return output_polygons


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.
from __future__ import annotations
import time
from ...mesh3d import Mesh3D
from ...geometry import GeoObject
from ...selection import FaceSelection, DomainSelection, EdgeSelection, Selection, encode_data
from ...physics.microwave.microwave_bc import PortBC, ModalPort
import numpy as np
import pyvista as pv
from typing import Iterable, Literal, Callable, Any
from ..display import BaseDisplay
from .display_settings import PVDisplaySettings
from matplotlib.colors import ListedColormap
from itertools import cycle
### Color scale

# Define the colors we want to use
col1 = np.array([57, 179, 227, 255])/255
col2 = np.array([22, 36, 125, 255])/255
col3 = np.array([33, 33, 33, 255])/255
col4 = np.array([173, 76, 7, 255])/255
col5 = np.array([250, 75, 148, 255])/255

cmap_names = Literal['bgy','bgyw','kbc','blues','bmw','bmy','kgy','gray','dimgray','fire','kb','kg','kr',
                     'bkr','bky','coolwarm','gwv','bjy','bwy','cwr','colorwheel','isolum','rainbow','fire',
                     'cet_fire','gouldian','kbgyw','cwr','CET_CBL1','CET_CBL3','CET_D1A']


def _gen_c_cycle():
    colors = [
        "#0000aa",
        "#aa0000",
        "#009900",
        "#990099",
        "#994400",
        "#005588"
    ]
    return cycle(colors)

C_CYCLE = _gen_c_cycle()

class _RunState:
    
    def __init__(self):
        self.state: bool = False
        self.ctr: int = 0
        
        
    def run(self):
        self.state = True
        self.ctr = 0
        
    def stop(self):
        self.state = False
        self.ctr = 0
        
    def step(self):
        self.ctr += 1
    
ANIM_STATE = _RunState()

def gen_cmap(mesh, N: int = 256):
    # build a linear grid of data‐values (not strictly needed for pure colormap)
    vmin, vmax = mesh['values'].min(), mesh['values'].max()
    mapping = np.linspace(vmin, vmax, N)
    
    # prepare output
    newcolors = np.empty((N, 4))
    
    # normalized positions of control points: start, middle, end
    control_pos = np.array([0.0, 0.25, 0.5, 0.75, 1]) * (vmax - vmin) + vmin
    # stack control colors
    controls = np.vstack([col1, col2, col3, col4, col5])
    
    # interp each RGBA channel independently
    for chan in range(4):
        newcolors[:, chan] = np.interp(mapping, control_pos, controls[:, chan])
    
    return ListedColormap(newcolors)

def setdefault(options: dict, **kwargs) -> dict:
    """Shorthand for overwriting non-existent keyword arguments with defaults

    Args:
        options (dict): The kwargs dict

    Returns:
        dict: the kwargs dict
    """
    for key in kwargs.keys():
        if options.get(key,None) is None:
            options[key] = kwargs[key]
    return options

def _logscale(dx, dy, dz):
    """
    Logarithmically scales vector magnitudes so that the largest remains unchanged
    and others are scaled down logarithmically.
    
    Parameters:
        dx, dy, dz (np.ndarray): Components of vectors.
        
    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: Scaled dx, dy, dz arrays.
    """
    dx = np.asarray(dx)
    dy = np.asarray(dy)
    dz = np.asarray(dz)

    # Compute original magnitudes
    mags = np.sqrt(dx**2 + dy**2 + dz**2)
    mags_nonzero = np.where(mags == 0, 1e-10, mags)  # avoid log(0)

    # Logarithmic scaling (scaled to max = original max)
    log_mags = np.log10(mags_nonzero)
    log_min = np.min(log_mags)
    log_max = np.max(log_mags)

    if log_max == log_min:
        # All vectors have the same length
        return dx, dy, dz

    # Normalize log magnitudes to [0, 1]
    log_scaled = (log_mags - log_min) / (log_max - log_min)

    # Scale back to original max magnitude
    max_mag = np.max(mags)
    new_mags = log_scaled * max_mag

    # Compute unit vectors
    unit_dx = dx / mags_nonzero
    unit_dy = dy / mags_nonzero
    unit_dz = dz / mags_nonzero

    # Apply scaled magnitudes
    scaled_dx = unit_dx * new_mags
    scaled_dy = unit_dy * new_mags
    scaled_dz = unit_dz * new_mags

    return scaled_dx, scaled_dy, scaled_dz

def _min_distance(xs, ys, zs):
    """
    Compute the minimum Euclidean distance between any two points
    defined by the 1D arrays xs, ys, zs.
    
    Parameters:
        xs (np.ndarray): x-coordinates of the points
        ys (np.ndarray): y-coordinates of the points
        zs (np.ndarray): z-coordinates of the points
    
    Returns:
        float: The minimum Euclidean distance between any two points
    """
    # Stack the coordinates into a (N, 3) array
    points = np.stack((xs, ys, zs), axis=-1)

    # Compute pairwise squared distances using broadcasting
    diff = points[:, np.newaxis, :] - points[np.newaxis, :, :]
    dists_squared = np.sum(diff ** 2, axis=-1)

    # Set diagonal to infinity to ignore zero distances to self
    np.fill_diagonal(dists_squared, np.inf)

    # Get the minimum distance
    min_dist = np.sqrt(np.min(dists_squared))
    return min_dist

def _norm(x, y, z):
    return np.sqrt(np.abs(x)**2 + np.abs(y)**2 + np.abs(z)**2)

def _select(obj: GeoObject | Selection) -> Selection:
    if isinstance(obj, GeoObject):
        return obj.select
    return obj

def _merge(lst: Iterable[GeoObject | Selection]) -> Selection:
    selections = [_select(item) for item in lst]
    dim = selections[0].dim
    all_tags = []
    for item in lst:
        all_tags.extend(_select(item).tags)
    
    if dim==1:
        return EdgeSelection(all_tags)
    elif dim==2:
        return FaceSelection(all_tags)
    elif dim==3:
        return DomainSelection(all_tags)
    else:
        return Selection(all_tags)

class _AnimObject:
    """ A private class containing the required information for plot items in a view
    that can be animated.
    """
    def __init__(self, 
                 field: np.ndarray,
                 T: Callable,
                 grid: pv.Grid,
                 actor: pv.Actor,
                 on_update: Callable):
        self.field: np.ndarray = field
        self.T: Callable = T
        self.grid: pv.Grid = grid
        self.actor: pv.Actor = actor
        self.on_update: Callable = on_update

    def update(self, phi: complex):
        self.on_update(self, phi)

class PVDisplay(BaseDisplay):

    def __init__(self, mesh: Mesh3D):
        self._mesh: Mesh3D = mesh
        self.set: PVDisplaySettings = PVDisplaySettings()
        
        # Animation options
        self._facetags: list[int] = []
        self._stop: bool = False
        self._objs: list[_AnimObject] = []
        self._do_animate: bool = False
        self._closed_via_x: bool = False
        self._Nsteps: int  = 0
        self._fps: int = 25
        self._ruler: ScreenRuler = ScreenRuler(self, 0.001)
        self._selector: ScreenSelector = ScreenSelector(self)
        self._stop = False
        self._objs = []

        self._plot = pv.Plotter()

        self._plot.add_key_event("m", self.activate_ruler) # type: ignore
        self._plot.add_key_event("f", self.activate_object) # type: ignore

        self._ctr: int = 0 
        
        self.camera_position = (1, -1, 1)     # +X, +Z, -Y
    
    def _wire_close_events(self):
        self._closed = False

        def mark_closed(*_):
            self._closed = True
            self._stop = True
        
        self._plot.add_key_event('q', lambda: mark_closed())
        
    def _update_camera(self):
        x,y,z = self._plot.camera.position
        d = (x**2+y**2+z**2)**(0.5)
        px, py, pz = self.camera_position
        dp = (px**2+py**2+pz**2)**(0.5)
        px, py, pz = px/dp, py/dp, pz/dp
        self._plot.camera.position = (d*px, d*py, d*pz)
        
    def activate_ruler(self):
        self._plot.disable_picking()
        self._selector.turn_off()
        self._ruler.toggle()

    def activate_object(self):
        self._plot.disable_picking()
        self._ruler.turn_off()
        self._selector.toggle()

    def show(self):
        """ Shows the Pyvista display. """
        self._ruler.min_length = max(1e-3, min(self._mesh.edge_lengths))
        self._update_camera()
        self._add_aux_items()
        # self._plot.renderer.enable_depth_peeling(20, 0.8)
        # self._plot.enable_anti_aliasing(self.set.anti_aliassing)
        if self._do_animate:
            self._wire_close_events()
            self.add_text('Press Q to close!',color='red', position='upper_left')
            self._plot.show(auto_close=False, interactive_update=True, before_close_callback=self._close_callback)
            self._animate()
        else:
            self._plot.show()
        
        self._reset()
    
    def set_mesh(self, mesh: Mesh3D):
        """Define the mesh to be used

        Args:
            mesh (Mesh3D): The mesh object
        """
        self._mesh = mesh

    def _reset(self):
        """ Resets key display parameters."""
        self._plot.close()
        self._plot = pv.Plotter()
        self._stop = False
        self._objs = []
        C_CYCLE = _gen_c_cycle()

    def _close_callback(self, arg):
        """The private callback function that stops the animation.
        """
        self._stop = True

    def _animate(self) -> None:
        """Private function that starts the animation loop.
        """
        
        self._stop = False

        # guard values
        steps = max(1, int(self._Nsteps))
        fps   = max(1, int(self._fps))
        dt    = 1.0 / fps
        next_tick = time.perf_counter()
        step = 0

        while (not self._stop
                and not self._closed_via_x
                and self._plot.render_window is not None):
            # process window/UI events so close button works
            self._plot.update()

            now = time.perf_counter()
            if now >= next_tick:
                step = (step + 1) % steps
                phi = np.exp(1j * (step / steps) * 2*np.pi)

                # update all animated objects
                for aobj in self._objs:
                    aobj.update(phi)

                # draw one frame
                self._plot.render()

                # schedule next frame; catch up if we fell behind
                next_tick += dt
                if now > next_tick + dt:
                    next_tick = now + dt

            # be kind to the CPU
            time.sleep(0.001)

        # ensure cleanup pathway runs once
        self._close_callback(None)

    def animate(self, Nsteps: int = 35, fps: int = 25) -> PVDisplay:
        """ Turns on the animation mode with the specified number of steps and FPS.

        All subsequent plot calls will automatically be animated. This method can be
        method chained.
        
        Args:
            Nsteps (int, optional): The number of frames in the loop. Defaults to 35.
            fps (int, optional): The number of frames per seocond, Defaults to 25

        Returns:
            PVDisplay: The same PVDisplay object

        Example:
        >>> display.animate().surf(...)
        >>> display.show()
        """
        self._Nsteps = Nsteps
        self._fps = fps
        self._do_animate = True
        return self
    
    ## CUSTOM METHODS
    def mesh_volume(self, volume: DomainSelection) -> pv.UnstructuredGrid:
        tets = self._mesh.get_tetrahedra(volume.tags)
        ntets = tets.shape[0]
        cells = np.zeros((ntets,5), dtype=np.int64)
        cells[:,1:] = self._mesh.tets[:,tets].T
        cells[:,0] = 4
        celltypes = np.full(ntets, fill_value=pv.CellType.TETRA, dtype=np.uint8)
        points = self._mesh.nodes.copy().T
        return pv.UnstructuredGrid(cells, celltypes, points)
    
    def _volume_edges(self, obj: GeoObject | Selection) -> pv.UnstructuredGrid:
        """Adds the edges of objects

        Args:
            obj (DomainSelection | None, optional): _description_. Defaults to None.

        Returns:
            pv.UnstructuredGrid: The unstrutured grid object
        """
        edge_ids = self._mesh.domain_edges(obj.dimtags)
        nedges = edge_ids.shape[0]
        cells = np.zeros((nedges,3), dtype=np.int64)
        cells[:,1:] = self._mesh.edges[:,edge_ids].T
        cells[:,0] = 2
        celltypes = np.full(nedges, fill_value=pv.CellType.CUBIC_LINE, dtype=np.uint8)
        points = self._mesh.nodes.copy().T
        return pv.UnstructuredGrid(cells, celltypes, points)
    
    def mesh_surface(self, surface: FaceSelection) -> pv.UnstructuredGrid:
        tris = self._mesh.get_triangles(surface.tags)
        ntris = tris.shape[0]
        cells = np.zeros((ntris,4), dtype=np.int64)
        cells[:,1:] = self._mesh.tris[:,tris].T
        cells[:,0] = 3
        celltypes = np.full(ntris, fill_value=pv.CellType.TRIANGLE, dtype=np.uint8)
        points = self._mesh.nodes.copy().T
        points[:,2] += self.set.z_boost
        return pv.UnstructuredGrid(cells, celltypes, points)
    
    def mesh(self, obj: GeoObject | Selection | Iterable) -> pv.UnstructuredGrid | None:
        if isinstance(obj, Iterable):
            obj = _merge(obj)
        else:
            obj = _select(obj)
        
        if isinstance(obj, DomainSelection):
            return self.mesh_volume(obj)
        elif isinstance(obj, FaceSelection):
            return self.mesh_surface(obj)
        else:
            return None

    ## OBLIGATORY METHODS
    def add_object(self, obj: GeoObject | Selection, mesh: bool = False, volume_mesh: bool = True, *args, **kwargs):
        
        show_edges = False
        opacity = obj.opacity
        line_width = 0.5
        color = obj.color_rgb
        metal = obj._metal
        style='surface'
        
        # Default render settings
        metallic = 0.05
        roughness = 0.5
        pbr = False
        
        if metal:
            pbr = True
            metallic = 0.8
            roughness = 0.3
        
        # Default keyword arguments when plotting Mesh mode.
        if mesh is True:
            show_edges = True
            opacity = 0.7
            line_width= 1
            style='wireframe'
            color=next(C_CYCLE)
        
        # Defining the default keyword arguments for PyVista
        kwargs = setdefault(kwargs, color=color, 
                            opacity=opacity, 
                            metallic=metallic, 
                            pbr=pbr,
                            roughness=roughness,
                            line_width=line_width, 
                            show_edges=show_edges, 
                            pickable=True, 
                            style=style)
        mesh_obj = self.mesh(obj)
        
        if mesh is True and volume_mesh is True:
            mesh_obj = mesh_obj.extract_all_edges()
        actor = self._plot.add_mesh(mesh_obj, *args, **kwargs)
        
        # Push 3D Geometries back to avoid Z-fighting with 2D geometries.
        if obj.dim==3:
            mapper = actor.GetMapper()
            mapper.SetResolveCoincidentTopology(1)
            mapper.SetRelativeCoincidentTopologyPolygonOffsetParameters(1,1)
        
        self._plot.add_mesh(self._volume_edges(_select(obj)), color='#000000', line_width=2, show_edges=True)

    def add_scatter(self, xs: np.ndarray, ys: np.ndarray, zs: np.ndarray):
        """Adds a scatter point cloud

        Args:
            xs (np.ndarray): The X-coordinate
            ys (np.ndarray): The Y-coordinate
            zs (np.ndarray): The Z-coordinate
        """
        cloud = pv.PolyData(np.array([xs,ys,zs]).T)
        self._plot.add_points(cloud)

    def add_portmode(self, port: PortBC, 
                     Npoints: int = 10, 
                     dv=(0,0,0), 
                     XYZ=None,
                     field: Literal['E','H'] = 'E', 
                     k0: float | None = None,
                     mode_number: int = 0) -> None:
        
        if XYZ:
            X,Y,Z = XYZ
        else:
            tris = self._mesh.get_triangles(port.selection.tags)
            X = self._mesh.tri_centers[0,tris]
            Y = self._mesh.tri_centers[1,tris]
            Z = self._mesh.tri_centers[2,tris]
        
        X = X+dv[0]
        Y = Y+dv[1]
        Z = Z+dv[2]
        xf = X.flatten()
        yf = Y.flatten()
        zf = Z.flatten()

        d = _min_distance(xf, yf, zf)

        if port.vintline is not None:
            xs, ys, zs = port.vintline.cpoint
            p_line = pv.Line(
                pointa=(xs[0], ys[0], zs[0]),
                pointb=(xs[-1], ys[-1], zs[-1]),
            )
            self._plot.add_mesh(
                p_line,
                color='red',
                pickable=False,
                line_width=3.0,
            )
        
        if k0 is None:
            if isinstance(port, ModalPort):
                k0 = port.get_mode(0).k0
            else:
                k0 = 1
        port.selected_mode = mode_number
        F = port.port_mode_3d_global(xf,yf,zf,k0, which=field)

        Fx = F[0,:].reshape(X.shape).T
        Fy = F[1,:].reshape(X.shape).T
        Fz = F[2,:].reshape(X.shape).T

        if field=='H':
            F = np.imag(F.T)
            Fnorm = np.sqrt(Fx.imag**2 + Fy.imag**2 + Fz.imag**2).T
        else:
            F = np.real(F.T)
            Fnorm = np.sqrt(Fx.real**2 + Fy.real**2 + Fz.real**2).T

        if XYZ is not None:
            grid = pv.StructuredGrid(X,Y,Z)
            self.add_surf(X,Y,Z,Fnorm, _fieldname = 'portfield')
            self._plot.add_mesh(grid, scalars = Fnorm.T, opacity=0.8, pickable=False)

        Emag = F/np.max(Fnorm.flatten())*d*3
        self._plot.add_arrows(np.array([xf,yf,zf]).T, Emag)

    def add_surf(self, 
                 x: np.ndarray,
                 y: np.ndarray,
                 z: np.ndarray,
                 field: np.ndarray,
                 scale: Literal['lin','log','symlog'] = 'lin',
                 cmap: cmap_names = 'coolwarm',
                 clim: tuple[float, float] | None = None,
                 opacity: float = 1.0,
                 symmetrize: bool = True,
                 _fieldname: str | None = None,
                 **kwargs,):
        """Add a surface plot to the display
        The X,Y,Z coordinates must be a 2D grid of data points. The field must be a real field with the same size.

        Args:
            x (np.ndarray): The X-grid array
            y (np.ndarray): The Y-grid array
            z (np.ndarray): The Z-grid array
            field (np.ndarray): The scalar field to display
            scale (Literal["lin","log","symlog"], optional): The colormap scaling¹. Defaults to 'lin'.
            cmap (cmap_names, optional): The colormap. Defaults to 'coolwarm'.
            clim (tuple[float, float], optional): Specific color limits (min, max). Defaults to None.
            opacity (float, optional): The opacity of the surface. Defaults to 1.0.
            symmetrize (bool, optional): Wether to force a symmetrical color limit (-A,A). Defaults to True.
        
        (¹): lin: f(x)=x, log: f(x)=log₁₀(|x|), symlog: f(x)=sgn(x)·log₁₀(1+|x·ln(10)|)
        """
        
        grid = pv.StructuredGrid(x,y,z)
        field_flat = field.flatten(order='F')


        if scale=='log':
            T = lambda x: np.log10(np.abs(x))
        elif scale=='symlog':
            T = lambda x: np.sign(x) * np.log10(1 + np.abs(x*np.log(10)))
        else:
            T = lambda x: x
        
        static_field = T(np.real(field_flat))
        if _fieldname is None:
            name = 'anim'+str(self._ctr)
        else:
            name = _fieldname
        self._ctr += 1
        grid[name] = static_field

        grid_no_nan = grid.threshold(scalars=name)

        # Determine color limits
        if clim is None:
            fmin = np.nanmin(static_field)
            fmax = np.nanmax(static_field)
            clim = (fmin, fmax)
        if symmetrize:
            lim = max(abs(clim[0]), abs(clim[1]))
            clim = (-lim, lim)

        kwargs = setdefault(kwargs, cmap=cmap, clim=clim, opacity=opacity, pickable=False, multi_colors=True)
        actor = self._plot.add_mesh(grid_no_nan, scalars=name, **kwargs)


        if self._do_animate:
            def on_update(obj: _AnimObject, phi: complex):
                field_anim = obj.T(np.real(obj.field * phi))
                obj.grid[name] = field_anim
            self._objs.append(_AnimObject(field_flat, T, grid_no_nan, actor, on_update))


    def add_title(self, title: str) -> None:
        """Adds a title

        Args:
            title (str): The title name
        """
        self._plot.add_text(
            title,
            position='upper_edge',
            font_size=18)

    def add_text(self, text: str, 
                 color: str = 'black', 
                 position: Literal['lower_left', 'lower_right', 'upper_left', 'upper_right', 'lower_edge', 'upper_edge', 'right_edge', 'left_edge']='upper_right',
                 abs_position: tuple[float, float, float] | None = None):
        viewport = False
        if abs_position is not None:
            final_position = abs_position
            viewport = True
        else:
            final_position = abs_position
        self._plot.add_text(
            text,
            position=final_position,
            color=color,
            font_size=18,
            viewport=viewport)
        
    def add_quiver(self, x: np.ndarray, y: np.ndarray, z: np.ndarray,
              dx: np.ndarray, dy: np.ndarray, dz: np.ndarray,
              scale: float = 1,
              color: tuple[float, float, float] | None = None,
              scalemode: Literal['lin','log'] = 'lin'):
        """Add a quiver plot to the display

        Args:
            x (np.ndarray): The X-coordinates
            y (np.ndarray): The Y-coordinates
            z (np.ndarray): The Z-coordinates
            dx (np.ndarray): The arrow X-magnitude
            dy (np.ndarray): The arrow Y-magnitude
            dz (np.ndarray): The arrow Z-magnitude
            scale (float, optional): The arrow scale. Defaults to 1.
            scalemode (Literal['lin','log'], optional): Wether to scale lin or log. Defaults to 'lin'.
        """
        x = x.flatten()
        y = y.flatten()
        z = z.flatten()
        dx = dx.flatten().real
        dy = dy.flatten().real
        dz = dz.flatten().real
        
        ids = np.invert(np.isnan(dx))
        
        x, y, z, dx, dy, dz = x[ids], y[ids], z[ids], dx[ids], dy[ids], dz[ids]
        
        dmin = _min_distance(x,y,z)

        dmax = np.max(_norm(dx,dy,dz))
        
        Vec = scale * np.array([dx,dy,dz]).T / dmax * dmin 
        Coo = np.array([x,y,z]).T
        if scalemode=='log':
            dx, dy, dz = _logscale(Vec[:,0], Vec[:,1], Vec[:,2])
            Vec[:,0] = dx
            Vec[:,1] = dy
            Vec[:,2] = dz
        
        kwargs = dict()
        if color is not None:
            kwargs['color'] = color
            
        pl = self._plot.add_arrows(Coo, Vec, scalars=None, clim=None, cmap=None, **kwargs)

        
    def add_contour(self,
                     X: np.ndarray,
                     Y: np.ndarray,
                     Z: np.ndarray,
                     V: np.ndarray,
                     Nlevels: int = 5,
                     symmetrize: bool = True,
                     cmap: str = 'viridis'):
        """Adds a 3D volumetric contourplot based on a 3D grid of X,Y,Z and field values


        Args:
            X (np.ndarray): A 3D Grid of X-values
            Y (np.ndarray): A 3D Grid of Y-values
            Z (np.ndarray): A 3D Grid of Z-values
            V (np.ndarray): The scalar quantity to plot ()
            Nlevels (int, optional): The number of contour levels. Defaults to 5.
            symmetrize (bool, optional): Wether to symmetrize the countour levels (-V,V). Defaults to True.
            cmap (str, optional): The color map. Defaults to 'viridis'.
        """
        Vf = V.flatten()
        vmin = np.min(np.real(Vf))
        vmax = np.max(np.real(Vf))
        if symmetrize:
            level = max(np.abs(vmin),np.abs(vmax))
            vmin, vmax = (-level, level)
        grid = pv.StructuredGrid(X,Y,Z)
        field = V.flatten(order='F')
        grid['anim'] = np.real(field)
        levels = list(np.linspace(vmin, vmax, Nlevels))
        contour = grid.contour(isosurfaces=levels)
        actor = self._plot.add_mesh(contour, opacity=0.25, cmap=cmap, pickable=False)

        if self._do_animate:
            def on_update(obj: _AnimObject, phi: complex):
                new_vals = np.real(obj.field * phi)
                obj.grid['anim'] = new_vals
                new_contour = obj.grid.contour(isosurfaces=levels)
                obj.actor.GetMapper().SetInputData(new_contour) # type: ignore
            
            self._objs.append(_AnimObject(field, lambda x: x, grid, actor, on_update)) # type: ignore

    def _add_aux_items(self) -> None:
        saved_camera = {
            "position": self._plot.camera.position,
            "focal_point": self._plot.camera.focal_point,
            "view_up": self._plot.camera.up,
            "view_angle": self._plot.camera.view_angle,
            "clipping_range": self._plot.camera.clipping_range
        }
        #self._plot.add_logo_widget('src/_img/logo.jpeg',position=(0.89,0.89), size=(0.1,0.1))    
        bounds = self._plot.bounds
        max_size = max([abs(dim) for dim in [bounds.x_max, bounds.x_min, bounds.y_max, bounds.y_min, bounds.z_max, bounds.z_min]])
        length = self.set.plane_ratio*max_size*2
        if self.set.draw_xplane:
            plane = pv.Plane(
                center=(0, 0, 0),
                direction=(1, 0, 0),    # normal vector pointing along +X
                i_size=length, # type: ignore
                j_size=length, # type: ignore
                i_resolution=1,
                j_resolution=1
            )
            self._plot.add_mesh(
                plane,
                color='red',
                opacity=self.set.plane_opacity,
                show_edges=False,
                pickable=False,
            )
            self._plot.add_mesh(
                plane,
                edge_opacity=1.0,
                edge_color='red',
                color='red',
                line_width=self.set.plane_edge_width,
                style='wireframe',
                pickable=False,
            )
            
        if self.set.draw_yplane:
            plane = pv.Plane(
                center=(0, 0, 0),
                direction=(0, 1, 0),    # normal vector pointing along +X
                i_size=length, # type: ignore
                j_size=length, # type: ignore
                i_resolution=1,
                j_resolution=1
            )
            self._plot.add_mesh(
                plane,
                color='green',
                opacity=self.set.plane_opacity,
                show_edges=False,
                pickable=False,
            )
            self._plot.add_mesh(
                plane,
                edge_opacity=1.0,
                edge_color='green',
                color='green',
                line_width=self.set.plane_edge_width,
                style='wireframe',
                pickable=False,
            )
        if self.set.draw_zplane:
            plane = pv.Plane(
                center=(0, 0, 0),
                direction=(0, 0, 1),    # normal vector pointing along +X
                i_size=length, # type: ignore
                j_size=length, # type: ignore
                i_resolution=1,
                j_resolution=1
            )
            self._plot.add_mesh(
                plane,
                color='blue',
                opacity=self.set.plane_opacity,
                show_edges=False,
                pickable=False,
            )
            self._plot.add_mesh(
                plane,
                edge_opacity=1.0,
                edge_color='blue',
                color='blue',
                line_width=self.set.plane_edge_width,
                style='wireframe',
                pickable=False,
            )
        # Draw X-axis
        if getattr(self.set, 'draw_xax', False):
            x_line = pv.Line(
                pointa=(-length, 0, 0),
                pointb=(length, 0, 0),
            )
            self._plot.add_mesh(
                x_line,
                color='red',
                line_width=self.set.axis_line_width,
                pickable=False,
            )

        # Draw Y-axis
        if getattr(self.set, 'draw_yax', False):
            y_line = pv.Line(
                pointa=(0, -length, 0),
                pointb=(0, length, 0),
            )
            self._plot.add_mesh(
                y_line,
                color='green',
                line_width=self.set.axis_line_width,
                pickable=False,
            )

        # Draw Z-axis
        if getattr(self.set, 'draw_zax', False):
            z_line = pv.Line(
                pointa=(0, 0, -length),
                pointb=(0, 0, length),
            )
            self._plot.add_mesh(
                z_line,
                color='blue',
                line_width=self.set.axis_line_width,
                pickable=False,
            )

        exponent = np.floor(np.log10(length))
        gs = 10 ** exponent
        N = np.ceil(length/gs)
        if N < 5:
            gs = gs/10
        L = (2*np.ceil(length/(2*gs))+1)*gs

        # XY grid at Z=0
        if self.set.show_zgrid:
            x_vals = np.arange(-L, L+gs, gs)
            y_vals = np.arange(-L, L+gs, gs)

            # lines parallel to X
            for y in y_vals:
                line = pv.Line(
                    pointa=(-L, y, 0),
                    pointb=(L, y, 0)
                )
                self._plot.add_mesh(line, color=self.set.grid_line_color, line_width=self.set.grid_line_width, opacity=0.5, edge_opacity=0.5,pickable=False)

            # lines parallel to Y
            for x in x_vals:
                line = pv.Line(
                    pointa=(x, -L, 0),
                    pointb=(x, L, 0)
                )
                self._plot.add_mesh(line, color=self.set.grid_line_color, line_width=self.set.grid_line_width, opacity=0.5, edge_opacity=0.5,pickable=False)


        # YZ grid at X=0
        if self.set.show_xgrid:
            y_vals = np.arange(-L, L+gs, gs)
            z_vals = np.arange(-L, L+gs, gs)

            # lines parallel to Y
            for z in z_vals:
                line = pv.Line(
                    pointa=(0, -L, z),
                    pointb=(0, L, z)
                )
                self._plot.add_mesh(line, color=self.set.grid_line_color, line_width=self.set.grid_line_width, opacity=0.5, edge_opacity=0.5, pickable=False)

            # lines parallel to Z
            for y in y_vals:
                line = pv.Line(
                    pointa=(0, y, -L),
                    pointb=(0, y, L)
                )
                self._plot.add_mesh(line, color=self.set.grid_line_color, line_width=self.set.grid_line_width, opacity=0.5, edge_opacity=0.5, pickable=False)


        # XZ grid at Y=0
        if self.set.show_ygrid:
            x_vals = np.arange(-L, L+gs, gs)
            z_vals = np.arange(-L, L+gs, gs)

            # lines parallel to X
            for z in z_vals:
                line = pv.Line(
                    pointa=(-length, 0, z),
                    pointb=(length, 0, z)
                )
                self._plot.add_mesh(line, color=self.set.grid_line_color, line_width=self.set.grid_line_width, opacity=0.5, edge_opacity=0.5, pickable=False)

            # lines parallel to Z
            for x in x_vals:
                line = pv.Line(
                    pointa=(x, 0, -length),
                    pointb=(x, 0, length)
                )
                self._plot.add_mesh(line, color=self.set.grid_line_color, line_width=self.set.grid_line_width, opacity=0.5, edge_opacity=0.5, pickable=False)

        if self.set.add_light:
            light = pv.Light()
            light.set_direction_angle(*self.set.light_angle) # type: ignore
            self._plot.add_light(light)

        self._plot.set_background(self.set.background_bottom, top=self.set.background_top) # type: ignore
        self._plot.add_axes() # type: ignore

        self._plot.camera.position = saved_camera["position"]
        self._plot.camera.focal_point = saved_camera["focal_point"]
        self._plot.camera.up = saved_camera["view_up"]
        self._plot.camera.view_angle = saved_camera["view_angle"]
        self._plot.camera.clipping_range = saved_camera["clipping_range"]

        

def freeze(function):

    def new_function(self, *args, **kwargs):
        cam = self.disp._plot.camera_position[:]
        self.disp._plot.suppress_rendering = True
        function(self, *args, **kwargs)
        self.disp._plot.camera_position = cam
        self.disp._plot.suppress_rendering = False
        self.disp._plot.render()
    return new_function


class ScreenSelector:

    def __init__(self, display: PVDisplay):
        self.disp: PVDisplay = display
        self.original_actors: list[pv.Actor] = []
        self.select_actors: list[pv.Actor] = []
        self.grids: list[pv.UnstructuredGrid] = []
        self.surfs: dict[int, np.ndarray] = dict()
        self.state = False

    def toggle(self):
        if self.state:
            self.turn_off()
        else:
            self.activate()

    def activate(self):
        self.original_actors = list(self.disp._plot.actors.values())

        for actor in self.original_actors:
            if isinstance(actor, pv.Text):
                continue
            actor.pickable = False
        
        if len(self.grids) == 0:
            for key in self.disp._facetags:
                tris = self.disp._mesh.get_triangles(key)
                ntris = tris.shape[0]
                cells = np.zeros((ntris,4), dtype=np.int64)
                cells[:,1:] = self.disp._mesh.tris[:,tris].T
                cells[:,0] = 3
                nodes = np.unique(self.disp._mesh.tris[:,tris].flatten())
                celltypes = np.full(ntris, fill_value=pv.CellType.TRIANGLE, dtype=np.uint8)
                points = self.disp._mesh.nodes.T
                grid = pv.UnstructuredGrid(cells, celltypes, points)
                grid._tag = key # type: ignore
                self.grids.append(grid)
                self.surfs[key] = points[nodes,:].T
        
        self.select_actors = []
        for grid in self.grids:
            actor = self.disp._plot.add_mesh(grid, opacity=0.001, color='red', pickable=True, name=f'FaceTag_{grid._tag}')
            self.select_actors.append(actor)

        def callback(actor: pv.Actor):
            key = int(actor.name.split('_')[1])
            points = self.surfs[key]
            xs = points[0,:]
            ys = points[1,:]
            zs = points[2,:]
            meanx = np.mean(xs)
            meany = np.mean(ys)
            meanz = np.mean(zs)
            data = (meanx, meany, meanz, min(xs), min(ys), min(zs), max(xs), max(ys), max(zs))
            encoded = encode_data(data) #type: ignore
            print(f'Face code key={key}: ', encoded)

        self.disp._plot.enable_mesh_picking(callback, style='surface', left_clicking=True, use_actor=True)
    
    def turn_off(self) -> None:
        for actor in self.select_actors:
            self.disp._plot.remove_actor(actor) # type: ignore
        self.select_actors = []
        for actor in self.original_actors:
            if isinstance(actor, pv.Text):
                continue
            actor.pickable = True

        
class ScreenRuler:

    def __init__(self, display: PVDisplay, min_length: float):
        self.disp: PVDisplay = display
        self.points: list[tuple] = [(0,0,0),(0,0,0)]
        self.text: pv.Text | None = None
        self.ruler: Any = None
        self.state: bool = False
        self.min_length: float = min_length
    
    @freeze
    def toggle(self):
        if not self.state:
            self.state = True
            self.disp._plot.enable_point_picking(self._add_point, left_clicking=True, tolerance=self.min_length)
        else:
            self.state = False
            self.disp._plot.disable_picking()

    @freeze
    def turn_off(self):
        self.state = False
        self.disp._plot.disable_picking()
    
    @property
    def dist(self) -> float:
        p1, p2 = self.points
        return ((p1[0]-p2[0])**2 + (p1[1]-p2[1])**2 + (p1[2]-p2[2])**2)**(0.5)
    
    @property
    def middle(self) -> tuple[float, float, float]:
        p1, p2 = self.points
        return ((p1[0]+p2[0])/2, (p1[1]+p2[1])/2, (p1[2]+p2[2])/2)
    
    @property
    def measurement_string(self) -> str:
        dist = self.dist
        p1, p2 = self.points
        dx = p2[0]-p1[0]
        dy = p2[1]-p1[1]
        dz = p2[2]-p1[2]
        return f'{dist*1000:.2f}mm (dx={1000.*dx:.4f}mm, dy={1000.*dy:.4f}mm, dz={1000.*dz:.4f}mm)'
    
    def set_ruler(self) -> None:
        if self.ruler is None:
            self.ruler = self.disp._plot.add_ruler(self.points[0], self.points[1], title=f'{1000*self.dist:.2f}mm') # type: ignore
        else:
            p1 = self.ruler.GetPositionCoordinate()
            p2 = self.ruler.GetPosition2Coordinate()
            p1.SetValue(*self.points[0])
            p2.SetValue(*self.points[1])
            self.ruler.SetTitle(f'{1000*self.dist:.2f}mm')
    
    @freeze
    def _add_point(self, point: tuple[float, float, float]):
        self.points = [point,self.points[0]]
        self.text = self.disp._plot.add_text(self.measurement_string, self.middle, name='RulerText')
        self.set_ruler()


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.
from numba import njit, f8, c16, i8, types # type: ignore
import numpy as np


@njit(i8[:](f8[:,:], i8[:,:], f8[:,:], i8[:]), cache=True, nogil=True)
def index_interp(coords: np.ndarray,
                tets: np.ndarray, 
                nodes: np.ndarray,
                tetids: np.ndarray):
    ''' Nedelec 2 tetrahedral interpolation of the analytic curl'''
    # Solution has shape (nEdges, nsols)
    nNodes = coords.shape[1]
    
    prop = np.full((nNodes, ), -1, dtype=np.int64)
    
    for i_iter in range(tetids.shape[0]):
        itet = tetids[i_iter]
        
        iv1, iv2, iv3, iv4 = tets[:, itet]

        v1 = nodes[:,iv1]
        v2 = nodes[:,iv2]
        v3 = nodes[:,iv3]
        v4 = nodes[:,iv4]

        bv1 = v2 - v1
        bv2 = v3 - v1
        bv3 = v4 - v1

        blocal = np.zeros((3,3))
        blocal[:,0] = bv1
        blocal[:,1] = bv2
        blocal[:,2] = bv3
        basis = np.linalg.pinv(blocal)

        coords_offset = coords - v1[:,np.newaxis]
        coords_local = (basis @ (coords_offset))


        inside = ((coords_local[0,:] + coords_local[1,:] + coords_local[2,:]) <= 1.00000001) & (coords_local[0,:] >= -1e-6) & (coords_local[1,:] >= -1e-6) & (coords_local[2,:] >= -1e-6)

        if inside.sum() == 0:
            continue
        
        prop[inside] = itet

    return prop



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
from ...simulation_data import BaseDataset, DataContainer
from ...elements.femdata import FEMBasis
from dataclasses import dataclass
import numpy as np
from typing import Literal, Callable
from loguru import logger
from .adaptive_freq import SparamModel
from ...cs import Axis, _parse_axis
from ...selection import FaceSelection
from ...geometry import GeoSurface
from ...mesh3d import Mesh3D
from ...const import Z0, MU0, EPS0

EMField = Literal[
    "er", "ur", "freq", "k0",
    "_Spdata", "_Spmapping", "_field", "_basis",
    "Nports", "Ex", "Ey", "Ez",
    "Hx", "Hy", "Hz",
    "mode", "beta",
]

def arc_on_plane(ref_dir, normal, angle_range_deg, num_points=100):
    """
    Generate theta/phi coordinates of an arc on a plane.

    Parameters
    ----------
    ref_dir : tuple (dx, dy, dz)
        Reference direction (angle zero) lying in the plane.
    normal : tuple (nx, ny, nz)
        Plane normal vector.
    angle_range_deg : tuple (deg_start, deg_end)
        Start and end angle of the arc in degrees.
    num_points : int
        Number of points along the arc.

    Returns
    -------
    theta : ndarray
        Array of theta angles (radians).
    phi : ndarray
        Array of phi angles (radians).
    """
    d = np.array(ref_dir, dtype=float)
    n = np.array(normal, dtype=float)

    # Normalize normal
    n = n / np.linalg.norm(n)

    # Project d into the plane
    d_proj = d - np.dot(d, n) * n
    if np.linalg.norm(d_proj) < 1e-12:
        raise ValueError("Reference direction is parallel to the normal vector.")

    e1 = d_proj / np.linalg.norm(d_proj)
    e2 = np.cross(n, e1)

    # Generate angles along the arc
    angles_deg = np.linspace(angle_range_deg[0], angle_range_deg[1], num_points)
    angles_rad = np.deg2rad(angles_deg)

    # Create unit vectors along the arc
    vectors = np.outer(np.cos(angles_rad), e1) + np.outer(np.sin(angles_rad), e2)

    # Convert to spherical angles
    ux, uy, uz = vectors[:,0], vectors[:,1], vectors[:,2]

    theta = np.arccos(uz)         # theta = arcsin(z)
    phi = np.arctan2(uy, ux)      # phi = atan2(y, x)

    return theta, phi

def renormalise_s(S: np.ndarray,
                  Zn: np.ndarray,
                  Z0: complex | float = 50) -> np.ndarray:
    S   = np.asarray(S,  dtype=complex)
    Zn  = np.asarray(Zn, dtype=complex)
    N   = S.shape[1]
    if S.shape[1:3] != (N, N):
        raise ValueError("S must have shape (M, N, N) with same N on both axes")
    if Zn.shape[1] != N:
        raise ValueError("Zn must be a length-N vector")

    # Constant matrices that do not depend on frequency
    
    W0_inv_sc = 1 / np.sqrt(Z0)               # scalar because Z0 is common
    I_N       = np.eye(N, dtype=complex)

    M = S.shape[0]
    S0 = np.empty_like(S)

    for k in range(M):
        Wref = np.diag(np.sqrt(Zn[k,:]))          # √Zn on the diagonal
        Sk = S[k, :, :]

        # Z  = Wref (I + S) (I – S)⁻¹ Wref
        Zk = Wref @ (I_N + Sk) @ np.linalg.inv(I_N - Sk) @ Wref

        # A  = W0⁻¹ Z W0⁻¹  → because W0 = √Z0·I → A = Z / Z0
        Ak = Zk * (W0_inv_sc ** 2)            # same as Zk / Z0

        # S0 = (A – I)(A + I)⁻¹
        S0[k, :, :] = (Ak - I_N) @ np.linalg.inv(Ak + I_N)

    return S0

def generate_ndim(
    outer_data: dict[str, list[float]],
    inner_data: list[float],
    outer_labels: tuple[str, ...]
) -> tuple[np.ndarray,...]:
    """
    Generates an N-dimensional grid of values from flattened data, and returns each axis array plus the grid.

    Parameters
    ----------
    outer_data : dict of {label: flat list of coordinates}
        Each key corresponds to one axis label, and the list contains coordinate values for each point.
    inner_data : list of float
        Flattened list of data values corresponding to each set of coordinates.
    outer_labels : tuple of str
        Order of axes (keys of outer_data) which defines the dimension order in the output array.

    Returns
    -------
    *axes : np.ndarray
        One 1D array for each axis, containing the sorted unique coordinates for that dimension, 
        in the order specified by outer_labels.
    grid : np.ndarray
        N-dimensional array of shape (n1, n2, ..., nN), where ni is the number of unique
        values along the i-th axis. Missing points are filled with np.nan.
    """
    # Convert inner data to numpy array
    values = np.asarray(inner_data)

    # Determine unique sorted coordinates for each axis
    axes = [np.unique(np.asarray(outer_data[label])) for label in outer_labels]
    grid_shape = tuple(axis.size for axis in axes)

    # Initialize grid with NaNs
    grid = np.full(grid_shape, np.nan, dtype=values.dtype)

    # Build coordinate arrays for each axis
    coords = [np.asarray(outer_data[label]) for label in outer_labels]

    # Map coordinates to indices in the grid for each axis
    idxs = [np.searchsorted(axes[i], coords[i]) for i in range(len(axes))]

    # Assign values into the grid
    grid[tuple(idxs)] = values

    # Return each axis array followed by the grid
    return (*axes, grid)

@dataclass
class Sparam:
    """
    S-parameter matrix indexed by arbitrary port/mode labels (ints or floats).
    Internally stores a square numpy array; externally uses your mapping
    to translate (port1, port2) → (i, j).
    """
    def __init__(self, port_nrs: list[int | float]) -> None:
        # build label → index map
        self.map: dict[int | float, int] = {label: idx 
                                            for idx, label in enumerate(port_nrs)}
        n = len(port_nrs)
        # zero‐initialize the S‐parameter matrix
        self.arry: np.ndarray = np.zeros((n, n), dtype=np.complex128)

    def get(self, port1: int | float, port2: int | float) -> complex:
        """
        Return the S-parameter S(port1, port2).
        Raises KeyError if either port1 or port2 is not in the mapping.
        """
        try:
            i = self.map[port1]
            j = self.map[port2]
        except KeyError as e:
            raise KeyError(f"Port/mode {e.args[0]!r} not found in mapping") from None
        return self.arry[i, j]

    def set(self, port1: int | float, port2: int | float, value: complex) -> None:
        """
        Set the S-parameter S(port1, port2) = value.
        Raises KeyError if either port1 or port2 is not in the mapping.
        """
        try:
            i = self.map[port1]
            j = self.map[port2]
        except KeyError as e:
            raise KeyError(f"Port/mode {e.args[0]!r} not found in mapping") from None
        self.arry[i, j] = value

    # allow S(param1, param2) → complex, as before
    def __call__(self, port1: int | float, port2: int | float) -> complex:
        return self.get(port1, port2)

    # allow array‐style access: S[1, 1] → complex
    def __getitem__(self, key: tuple[int | float, int | float]) -> complex:
        port1, port2 = key
        return self.get(port1, port2)

    # allow array‐style setting: S[1, 2] = 0.3 + 0.1j
    def __setitem__(
        self,
        key: tuple[int | float, int | float],
        value: complex
    ) -> None:
        port1, port2 = key
        self.set(port1, port2, value)

@dataclass
class PortProperties:
    port_number: int = -1
    k0: float | None= None
    beta: float | None = None
    Z0: float | complex | None = None
    Pout: float | None = None
    mode_number: int = 1

class MWData:
    scalar: BaseDataset[MWScalar, MWScalarNdim]
    field:   BaseDataset[MWField, None]

    def __init__(self):
        self.scalar = BaseDataset[MWScalar, MWScalarNdim](MWScalar, MWScalarNdim, True)
        self.field = BaseDataset[MWField, None](MWField, None, False)
        self.sim: DataContainer = DataContainer()

    def setreport(self, report, **vars):
        self.sim.new(**vars)['report'] = report
    
@dataclass
class FarFieldData:
    E: np.ndarray
    H: np.ndarray
    theta: np.ndarray
    phi: np.ndarray
    ang: np.ndarray | None = None

    @property
    def Ex(self) -> np.ndarray:
        return self.E[0,:]
    
    @property
    def Ey(self) -> np.ndarray:
        return self.E[1,:]
    
    @property
    def Ez(self) -> np.ndarray:
        return self.E[2,:]
    
    @property
    def Hx(self) -> np.ndarray:
        return self.H[0,:]
    
    @property
    def Hy(self) -> np.ndarray:
        return self.H[1,:]
    
    @property
    def Hz(self) -> np.ndarray:
        return self.H[2,:]
    
    @property
    def Etheta(self) -> np.ndarray:
        thx = np.cos(self.theta)*np.cos(self.phi)
        thy = np.cos(self.theta)*np.sin(self.phi)
        thz = -np.sin(self.theta)
        return thx*self.E[0,:] + thy*self.E[1,:] + thz*self.E[2,:]
    
    @property
    def Ephi(self) -> np.ndarray:
        phx = -np.sin(self.phi)
        phy = np.cos(self.phi)
        phz = np.zeros_like(self.theta)
        return phx*self.E[0,:] + phy*self.E[1,:] + phz*self.E[2,:]
    
    @property
    def Erhcp(self) -> np.ndarray:
        return (self.Etheta + 1j*self.Ephi)/np.sqrt(2)
    
    @property
    def Elhcp(self) -> np.ndarray:
        return (self.Etheta - 1j*self.Ephi)/np.sqrt(2)
    
    @property
    def AR(self) -> np.ndarray:
        R = np.abs(self.Erhcp)
        L = np.abs(self.Elhcp)
        return (R+L)/(R-L)
    
    @property
    def normE(self) -> np.ndarray:
        return np.sqrt(np.abs(self.E[0,:])**2 + np.abs(self.E[1,:])**2 + np.abs(self.E[2,:])**2)
    
    @property
    def normH(self) -> np.ndarray:
        return np.sqrt(np.abs(self.H[0,:])**2 + np.abs(self.H[1,:])**2 + np.abs(self.H[2,:])**2)
    
    
    def surfplot(self, 
             polarization: Literal['Ex','Ey','Ez','Etheta','Ephi','normE','Erhcp','Elhcp','AR'],
             quantity: Literal['abs','real','imag','angle'] = 'abs',
             isotropic: bool = True, dB: bool = False, dBfloor: float = -30, rmax: float | None = None,
             offset: tuple[float, float, float] = (0,0,0)) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Returns the parameters to be used as positional arguments for the display.add_surf() function.

        Example:
        >>> model.display.add_surf(*dataset.field[n].farfield_3d(...).surfplot())

        Args:
            polarization ('Ex','Ey','Ez','Etheta','Ephi','normE'): What quantity to plot
            isotropic (bool, optional): Whether to look at the ratio with isotropic antennas. Defaults to True.
            dB (bool, optional): Whether to plot in dB's. Defaults to False.
            dBfloor (float, optional): The dB value to take as R=0. Defaults to -10.

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: The X, Y, Z, F values
        """
        fmap = {
            'abs': np.abs,
            'real': np.real,
            'imag': np.imag,
            'angle': np.angle,
        }
        mapping = fmap.get(quantity.lower(),np.abs)
        
        F = mapping(self.__dict__.get(polarization, self.normE))
        
        if isotropic:
            F = F/np.sqrt(Z0/(2*np.pi))
        if dB:
            F = 20*np.log10(np.clip(np.abs(F), a_min=10**(dBfloor/20), a_max = 1e9))-dBfloor
        if rmax is not None:
            F = rmax * F/np.max(F)
        xs = F*np.sin(self.theta)*np.cos(self.phi) + offset[0]
        ys = F*np.sin(self.theta)*np.sin(self.phi) + offset[1]
        zs = F*np.cos(self.theta) + offset[2]

        return xs, ys, zs, F

@dataclass
class EHField:
    x: np.ndarray
    y: np.ndarray
    z: np.ndarray
    Ex: np.ndarray
    Ey: np.ndarray
    Ez: np.ndarray
    Hx: np.ndarray
    Hy: np.ndarray
    Hz: np.ndarray
    freq: float
    er: np.ndarray
    ur: np.ndarray

    @property
    def k0(self) -> float:
        return self.freq*2*np.pi/299792458
    
    @property
    def Px(self) -> np.ndarray:
        return EPS0*(self.er-1)*self.Ex
    
    @property
    def Py(self) -> np.ndarray:
        return EPS0*(self.er-1)*self.Ey
    
    @property
    def Pz(self) -> np.ndarray:
        return EPS0*(self.er-1)*self.Ez
    
    @property
    def Dx(self) -> np.ndarray:
        return self.Ex*self.er
    
    @property
    def Dy(self) -> np.ndarray:
        return self.Ey*self.er
    
    @property
    def Dz(self) -> np.ndarray:
        return self.Ez*self.er
    
    @property
    def Bx(self) -> np.ndarray:
        return self.Hx/self.ur
    
    @property
    def By(self) -> np.ndarray:
        return self.Hy/self.ur
    
    @property
    def Bz(self) -> np.ndarray:
        return self.Hz/self.ur
    
    @property
    def Emat(self) -> np.ndarray:
        return np.array([self.Ex, self.Ey, self.Ez])
    
    @property
    def Hmat(self) -> np.ndarray:
        return np.array([self.Hx, self.Hy, self.Hz])
    
    @property
    def Pmat(self) -> np.ndarray:
        return np.array([self.Px, self.Py, self.Pz])
    
    @property
    def Bmat(self) -> np.ndarray:
        return np.array([self.Bx, self.By, self.Bz])
    
    @property
    def Dmat(self) -> np.ndarray:
        return np.array([self.Dx, self.Dy, self.Dz])
    
    @property
    def EH(self) -> tuple[np.ndarray, np.ndarray]:
        ''' Return the electric and magnetic field as a tuple of numpy arrays '''
        return np.array([self.Ex, self.Ey, self.Ez]), np.array([self.Hx, self.Hy, self.Hz])
    
    @property
    def E(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        ''' Return the electric field as a tuple of numpy arrays '''
        return self.Ex, self.Ey, self.Ez
    
    @property
    def Sx(self) -> np.ndarray:
        return self.Ey*self.Hz - self.Ez*self.Hy
    
    @property
    def Sy(self) -> np.ndarray:
        return self.Ez*self.Hx - self.Ex*self.Hz
    
    @property
    def Sz(self) -> np.ndarray:
        return self.Ex*self.Hy - self.Ey*self.Hx
    
    @property
    def B(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        ''' Return the magnetic field as a tuple of numpy arrays '''
        return self.Bx, self.By, self.Bz
    
    @property
    def P(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        ''' Return the polarization field as a tuple of numpy arrays '''
        return self.Px, self.Py, self.Pz

    @property
    def D(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        ''' Return the electric displacement field as a tuple of numpy arrays '''
        return self.Bx, self.By, self.Bz
    
    @property
    def H(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        ''' Return the magnetic field as a tuple of numpy arrays '''
        return self.Hx, self.Hy, self.Hz

    @property
    def S(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        ''' Return the poynting vector field as a tuple of numpy arrays '''
        return self.Sx, self.Sy, self.Sz
    
    @property
    def normE(self) -> np.ndarray:
        """The complex norm of the E-field
        """
        return np.sqrt(np.abs(self.Ex)**2 + np.abs(self.Ey)**2 + np.abs(self.Ez)**2)
    
    @property
    def normH(self) -> np.ndarray:
        """The complex norm of the H-field"""
        return np.sqrt(np.abs(self.Hx)**2 + np.abs(self.Hy)**2 + np.abs(self.Hz)**2)
    
    @property
    def normP(self) -> np.ndarray:
        """The complex norm of the P-field
        """
        return np.sqrt(np.abs(self.Px)**2 + np.abs(self.Py)**2 + np.abs(self.Pz)**2)
    
    @property
    def normB(self) -> np.ndarray:
        """The complex norm of the B-field
        """
        return np.sqrt(np.abs(self.Bx)**2 + np.abs(self.By)**2 + np.abs(self.Bz)**2)
    
    @property
    def normD(self) -> np.ndarray:
        """The complex norm of the D-field
        """
        return np.sqrt(np.abs(self.Dx)**2 + np.abs(self.Dy)**2 + np.abs(self.Dz)**2)
    
    @property
    def normS(self) -> np.ndarray:
        """The complex norm of the S-field
        """
        return np.sqrt(np.abs(self.Sx)**2 + np.abs(self.Sy)**2 + np.abs(self.Sz)**2)
    
    def vector(self, field: Literal['E','H'], metric: Literal['real','imag','complex'] = 'real') -> tuple[np.ndarray, np.ndarray,np.ndarray,np.ndarray,np.ndarray,np.ndarray]:
        """Returns the X,Y,Z,Fx,Fy,Fz data to be directly cast into plot functions.

        The field can be selected by a string literal. The metric of the complex vector field by the metric.
        For animations, make sure to always use the complex metric.

        Args:
            field ('E','H'): The field to return
            metric ([]'real','imag','complex'], optional): the metric to impose on the field. Defaults to 'real'.

        Returns:
            tuple[np.ndarray,...]: The X,Y,Z,Fx,Fy,Fz arrays
        """
        Fx, Fy, Fz = getattr(self, field)
        
        if metric=='real':
            Fx, Fy, Fz = Fx.real, Fy.real, Fz.real
        elif metric=='imag':
            Fx, Fy, Fz = Fx.imag, Fy.imag, Fz.imag
        
        return self.x, self.y, self.z, Fx, Fy, Fz
    
    def scalar(self, field: Literal['Ex','Ey','Ez','Hx','Hy','Hz','normE','normH'], metric: Literal['abs','real','imag','complex'] = 'real') -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Returns the data X, Y, Z, Field based on the interpolation

        For animations, make sure to select the complex metric.

        Args:
            field (str): The field to plot
            metric (str, optional): The metric to impose on the plot. Defaults to 'real'.

        Returns:
            (X,Y,Z,Field): The coordinates plus field scalar
        """
        field_arry = getattr(self, field)
        if metric=='abs':
            field = np.abs(field_arry)
        elif metric=='real':
            field = field_arry.real
        elif metric=='imag':
            field = field_arry.imag
        elif metric=='complex':
            field = field_arry
        return self.x, self.y, self.z, field_arry

class _EHSign:
    """A small class to manage the sign of field components when computing the far-field with Stratton-Chu
    """
    def __init__(self):
        self.Ex = 1
        self.Ey = 1
        self.Ez = 1
        self.Hx = 1
        self.Hy = 1
        self.Hz = 1

    def fE(self):
        self.Ex = -1*self.Ex
        self.Ey = -1*self.Ey
        self.Ez = -1*self.Ez

    def fH(self):
        self.Hx = -1*self.Hx
        self.Hy = -1*self.Hy
        self.Hz = -1*self.Hz

    def fX(self):
        self.Ex = -1*self.Ex
        self.Hx = -1*self.Hx

    def fY(self):
        self.Ey = -1*self.Ey
        self.Hy = -1*self.Hy

    def fZ(self):
        self.Ez = -1*self.Ez
        self.Hz = -1*self.Hz

    def apply(self, symmetry: str):
        f, c = symmetry
        if f=='E':
            self.fE()
        elif f=='H':
            self.fH()

        if c=='x':
            self.fX()
        elif c=='y':
            self.fY()
        elif c=='z':
            self.fZ()
        
    def flip_field(self, E: tuple, H: tuple):
        Ex, Ey, Ez = E
        Hx, Hy, Hz = H
        return (Ex*self.Ex, Ey*self.Ey, Ez*self.Ez), (Hx*self.Hx, Hy*self.Hy, Hz*self.Hz)
    
class MWField:
    
    def __init__(self):
        self._der: np.ndarray = None
        self._dur: np.ndarray = None
        self.freq: float = None
        self.basis: FEMBasis = None
        self._fields: dict[int, np.ndarray] = dict()
        self._mode_field: np.ndarray = None
        self.excitation: dict[int, complex] = dict()
        self.Nports: int = None
        self.port_modes: list[PortProperties] = []
        self.Ex: np.ndarray = None
        self.Ey: np.ndarray = None
        self.Ez: np.ndarray = None
        self.Hx: np.ndarray = None
        self.Hy: np.ndarray = None
        self.Hz: np.ndarray = None
        self.er: np.ndarray = None
        self.ur: np.ndarray = None

    def add_port_properties(self, 
                            port_number: int,
                            mode_number: int,
                            k0: float,
                            beta: float,
                            Z0: float | complex | None,
                            Pout: float) -> None:
        self.port_modes.append(PortProperties(port_number=port_number,
                                              mode_number=mode_number,
                                              k0 = k0,
                                              beta=beta,
                                              Z0=Z0,
                                              Pout=Pout))
    
    @property
    def mesh(self) -> Mesh3D:
        return self.basis.mesh
    
    @property
    def k0(self) -> float:
        return self.freq*2*np.pi/299792458
    
    @property
    def _field(self) -> np.ndarray:
        if self._mode_field is not None:
            return self._mode_field
        return sum([self.excitation[mode.port_number]*self._fields[mode.port_number] for mode in self.port_modes]) # type: ignore
    
    def set_field_vector(self) -> None:
        """Defines the default excitation coefficients for the current dataset"""
        self.excitation = {key: 0.0 for key in self._fields.keys()}
        self.excitation[self.port_modes[0].port_number] = 1.0 + 0j

    @property
    def EH(self) -> tuple[np.ndarray, np.ndarray]:
        ''' Return the electric and magnetic field as a tuple of numpy arrays '''
        return np.array([self.Ex, self.Ey, self.Ez]), np.array([self.Hx, self.Hy, self.Hz])
    
    @property
    def E(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        ''' Return the electric field as a tuple of numpy arrays '''
        return self.Ex, self.Ey, self.Ez
    
    @property
    def normE(self) -> np.ndarray:
        """The complex norm of the E-field
        """
        return np.sqrt(np.abs(self.Ex)**2 + np.abs(self.Ey)**2 + np.abs(self.Ez)**2)
    
    @property
    def normH(self) -> np.ndarray:
        """The complex norm of the H-field"""
        return np.sqrt(np.abs(self.Hx)**2 + np.abs(self.Hy)**2 + np.abs(self.Hz)**2)
    
    @property
    def Emat(self) -> np.ndarray:
        return np.array([self.Ex, self.Ey, self.Ez])
    
    @property
    def Hmat(self) -> np.ndarray:
        return np.array([self.Hx, self.Hy, self.Hz])

    @property
    def H(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        ''' Return the magnetic field as a tuple of numpy arrays '''
        return self.Hx, self.Hy, self.Hz
    
    def interpolate(self, xs: np.ndarray, ys: np.ndarray, zs: np.ndarray) -> EHField:
        ''' Interpolate the dataset in the provided xs, ys, zs values'''
        if isinstance(xs, (float, int, complex)):
            xs = np.array([xs,])
            ys = np.array([ys,])
            zs = np.array([zs,])
            
        shp = xs.shape
        xf = xs.flatten()
        yf = ys.flatten()
        zf = zs.flatten()
        Ex, Ey, Ez = self.basis.interpolate(self._field, xf, yf, zf)
        self.Ex = Ex.reshape(shp)
        self.Ey = Ey.reshape(shp)
        self.Ez = Ez.reshape(shp)

        
        constants = 1/ (-1j*2*np.pi*self.freq*(self._dur*MU0) )
        Hx, Hy, Hz = self.basis.interpolate_curl(self._field, xf, yf, zf, constants)
        ids = self.basis.interpolate_index(xf, yf, zf)
        self.er = self._der[ids].reshape(shp)
        self.ur = self._dur[ids].reshape(shp)
        self.Hx = Hx.reshape(shp)
        self.Hy = Hy.reshape(shp)
        self.Hz = Hz.reshape(shp)

        self._x = xs
        self._y = ys
        self._z = zs
        return EHField(xs, ys, zs, self.Ex, self.Ey, self.Ez, self.Hx, self.Hy, self.Hz, self.freq, self.er, self.ur)

    def cutplane(self, 
                     ds: float,
                     x: float | None = None,
                     y: float | None = None,
                     z: float | None = None) -> EHField:
        """Create a cartesian cut plane (XY, YZ or XZ) and compute the E and H-fields there

        Only one coordiante and thus cutplane may be defined. If multiple are defined only the last (x->y->z) is used.
        
        Args:
            ds (float): The discretization step size
            x (float | None, optional): The X-coordinate in case of a YZ-plane. Defaults to None.
            y (float | None, optional): The Y-coordinate in case of an XZ-plane. Defaults to None.
            z (float | None, optional): The Z-coordinate in case of an XY-plane. Defaults to None.

        Returns:
            EHField: The resultant EHField object
        """
        xb, yb, zb = self.basis.bounds
        xs = np.linspace(xb[0], xb[1], int((xb[1]-xb[0])/ds))
        ys = np.linspace(yb[0], yb[1], int((yb[1]-yb[0])/ds))
        zs = np.linspace(zb[0], zb[1], int((zb[1]-zb[0])/ds))
        
        if x is not None:
            Y,Z = np.meshgrid(ys, zs)
            X = x*np.ones_like(Y)
        if y is not None:
            X,Z = np.meshgrid(xs, zs)
            Y = y*np.ones_like(X)
        if z is not None:
            X,Y = np.meshgrid(xs, ys)
            Z = z*np.ones_like(Y)
        return self.interpolate(X,Y,Z)
    
    def cutplane_normal(self,
             point=(0,0,0),
             normal=(0,0,1),
             npoints: int = 300) -> EHField:
        """
        Take a 2D slice of the field along an arbitrary plane.
        Args:
            point: (x0,y0,z0), a point on the plane
            normal: (nx,ny,nz), plane normal vector
            npoints: number of grid points per axis
        """

        n = np.array(normal, dtype=float)
        n /= np.linalg.norm(n)
        point = np.array(point) 

        tmp = np.array([1,0,0]) if abs(n[0]) < 0.9 else np.array([0,1,0])
        u = np.cross(n, tmp)
        u /= np.linalg.norm(u)
        v = np.cross(n, u)
        
        xb, yb, zb = self.basis.bounds
        nx, ny, nz = 5, 5, 5
        Xg = np.linspace(xb[0], xb[1], nx)
        Yg = np.linspace(yb[0], yb[1], ny)
        Zg = np.linspace(zb[0], zb[1], nz)
        Xg, Yg, Zg = np.meshgrid(Xg, Yg, Zg, indexing='ij')
        geometry = np.vstack([Xg.ravel(), Yg.ravel(), Zg.ravel()]).T  # Nx3
        
        rel_pts = geometry - point
        S = rel_pts @ u
        T = rel_pts @ v 
        
        margin = 0.01
        s_min, s_max = S.min(), S.max()
        t_min, t_max = T.min(), T.max()
        s_bounds = (s_min - margin*(s_max-s_min), s_max + margin*(s_max-s_min))
        t_bounds = (t_min - margin*(t_max-t_min), t_max + margin*(t_max-t_min))

        S_grid = np.linspace(s_bounds[0], s_bounds[1], npoints)
        T_grid = np.linspace(t_bounds[0], t_bounds[1], npoints)
        S_mesh, T_mesh = np.meshgrid(S_grid, T_grid)

        X = point[0] + S_mesh*u[0] + T_mesh*v[0]
        Y = point[1] + S_mesh*u[1] + T_mesh*v[1]
        Z = point[2] + S_mesh*u[2] + T_mesh*v[2]

        return self.interpolate(X, Y, Z)
    
    
    def grid(self, ds: float) -> EHField:
        """Interpolate a uniform grid sampled at ds

        Args:
            ds (float): the sampling distance

        Returns:
            This object
        """
        xb, yb, zb = self.basis.bounds
        xs = np.linspace(xb[0], xb[1], int((xb[1]-xb[0])/ds))
        ys = np.linspace(yb[0], yb[1], int((yb[1]-yb[0])/ds))
        zs = np.linspace(zb[0], zb[1], int((zb[1]-zb[0])/ds))
        X, Y, Z = np.meshgrid(xs, ys, zs)
        return self.interpolate(X,Y,Z)
    
    def vector(self, field: Literal['E','H'], metric: Literal['real','imag','complex'] = 'real') -> tuple[np.ndarray, np.ndarray,np.ndarray,np.ndarray,np.ndarray,np.ndarray]:
        """Returns the X,Y,Z,Fx,Fy,Fz data to be directly cast into plot functions.

        The field can be selected by a string literal. The metric of the complex vector field by the metric.
        For animations, make sure to always use the complex metric.

        Args:
            field ('E','H'): The field to return
            metric ([]'real','imag','complex'], optional): the metric to impose on the field. Defaults to 'real'.

        Returns:
            tuple[np.ndarray,...]: The X,Y,Z,Fx,Fy,Fz arrays
        """
        if field=='E':
            Fx, Fy, Fz = self.Ex, self.Ey, self.Ez
        elif field=='H':
            Fx, Fy, Fz = self.Hx, self.Hy, self.Hz
        
        if metric=='real':
            Fx, Fy, Fz = Fx.real, Fy.real, Fz.real
        elif metric=='imag':
            Fx, Fy, Fz = Fx.imag, Fy.imag, Fz.imag
        
        return self._x, self._y, self._z, Fx, Fy, Fz
    
    def scalar(self, field: Literal['Ex','Ey','Ez','Hx','Hy','Hz','normE','normH'], metric: Literal['abs','real','imag','complex'] = 'real') -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Returns the data X, Y, Z, Field based on the interpolation

        For animations, make sure to select the complex metric.

        Args:
            field (str): The field to plot
            metric (str, optional): The metric to impose on the plot. Defaults to 'real'.

        Returns:
            (X,Y,Z,Field): The coordinates plus field scalar
        """
        field = getattr(self, field)
        if metric=='abs':
            field = np.abs(field)
        elif metric=='real':
            field = field.real
        elif metric=='imag':
            field = field.imag
        elif metric=='complex':
            field = field
        return self._x, self._y, self._z, field
    
    def farfield_2d(self,ref_direction: tuple[float,float,float] | Axis,
                         plane_normal: tuple[float,float,float] | Axis,
                         faces: FaceSelection | GeoSurface,
                         ang_range: tuple[float, float] = (-180, 180),
                         Npoints: int = 201,
                         origin: tuple[float, float, float] | None = None,
                         syms: list[Literal['Ex','Ey','Ez', 'Hx','Hy','Hz']] | None = None) -> FarFieldData:#tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Compute the farfield electric and magnetic field defined by a circle.

        Args:
            ref_direction (tuple[float,float,float] | Axis): The direction for angle=0
            plane_normal (tuple[float,float,float] | Axis): The rotation axis of the angular cutplane
            faces (FaceSelection | GeoSurface): The faces to integrate over
            ang_range (tuple[float, float], optional): The angular rage limits. Defaults to (-180, 180).
            Npoints (int, optional): The number of angular points. Defaults to 201.
            origin (tuple[float, float, float], optional): The farfield origin. Defaults to (0,0,0).
            syms (list[Literal['Ex','Ey','Ez','Hx','Hy','Hz']], optional): E and H-plane symmetry planes where Ex is E-symmetry in x=0. Defaults to []

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: Angles (N,), E(3,N), H(3,N)
        """
        refdir = _parse_axis(ref_direction).np
        plane_normal_parsed = _parse_axis(plane_normal).np
        theta, phi = arc_on_plane(refdir, plane_normal_parsed, ang_range, Npoints)
        E,H = self.farfield(theta, phi, faces, origin, syms = syms)
        angs = np.linspace(*ang_range, Npoints)*np.pi/180
        return FarFieldData(E, H, theta, phi, ang=angs)

    def farfield_3d(self, 
                    faces: FaceSelection | GeoSurface,
                    thetas: np.ndarray | None = None,
                    phis: np.ndarray | None = None,
                    origin: tuple[float, float, float] | None = None,
                    syms: list[Literal['Ex','Ey','Ez', 'Hx','Hy','Hz']] | None = None) -> FarFieldData:
        """Compute the farfield in a 3D angular grid

        If thetas and phis are not provided, they default to a sample space of 2 degrees.

        Args:
            faces (FaceSelection | GeoSurface): The integration faces
            thetas (np.ndarray, optional): The 1D array of theta values. Defaults to None.
            phis (np.ndarray, optional): A 1D array of phi values. Defaults to None.
            origin (tuple[float, float, float], optional): The boundary normal alignment origin. Defaults to (0,0,0).
            syms (list[Literal['Ex','Ey','Ez','Hx','Hy','Hz']], optional): E and H-plane symmetry planes where Ex is E-symmetry in x=0. Defaults to []
        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: The 2D theta, phi, E and H matrices.
        """
        if thetas is None:
            thetas = np.linspace(0,np.pi, 91)
        if phis is None:
            phis = np.linspace(-np.pi, np.pi, 181)

        T,P = np.meshgrid(thetas, phis)

        E, H = self.farfield(T.flatten(), P.flatten(), faces, origin, syms=syms)
        E = E.reshape((3, ) + T.shape)
        H = H.reshape((3, ) + T.shape)
        
        return FarFieldData(E, H, T, P)

    def farfield(self, theta: np.ndarray,
                 phi: np.ndarray,
                 faces: FaceSelection | GeoSurface,
                 origin: tuple[float, float, float] | None = None,
                 syms: list[Literal['Ex','Ey','Ez', 'Hx','Hy','Hz']] | None = None) -> tuple[np.ndarray, np.ndarray]:
        """Compute the farfield at the provided theta/phi coordinates

        Args:
            theta (np.ndarray): The Theta coordinates as (N,) 1D Array
            phi (np.ndarray): The Phi coordinates as (N,) 1D Array
            faces (FaceSelection | GeoSurface): the faces to use as integration boundary
            origin (tuple[float, float, float], optional): The surface normal origin. Defaults to (0,0,0).
            syms (list[Literal['Ex','Ey','Ez','Hx','Hy','Hz']], optional): E and H-plane symmetry planes where Ex is E-symmetry in x=0. Defaults to []

        Returns:
            tuple[np.ndarray, np.ndarray]: The E and H field as (3,N) arrays
        """
        if syms is None:
            syms = []

        from .sc import stratton_chu
        surface = self.basis.mesh.boundary_surface(faces.tags, origin)
        field = self.interpolate(*surface.exyz)
        
        Eff,Hff = stratton_chu(field.E, field.H, surface, theta, phi, self.k0)
        
        if len(syms)==0:
            return Eff, Hff

        if len(syms)==1:
            perms = ((syms[0], '##', '##'),)
        elif len(syms)==2:
            s1, s2 = syms
            perms = ((s1, '##', '##'), (s2, '##', '##'), (s1, s2, '##'))
        elif len(syms)==3:
            s1, s2, s3 = syms
            perms = ((s1, '##', '##'), (s2, '##', '##'), (s3, '##', '##'), (s1, s2, '##'), (s1, s3, '##'), (s2, s3, '##'), (s1, s2, s3))
        
        for s1, s2, s3 in perms:
            surf = surface.copy()
            ehf = _EHSign()
            ehf.apply(s1)
            ehf.apply(s2)
            ehf.apply(s3)
            Ef, Hf = ehf.flip_field(field.E, field.H)
            surf.flip(s1[1])
            surf.flip(s2[1])
            surf.flip(s3[1])
            E2, H2 = stratton_chu(Ef, Hf, surf, theta, phi, self.k0)
            Eff = Eff + E2
            Hff = Hff + H2
        
        return Eff, Hff

    def optycal_surface(self, faces: FaceSelection | GeoSurface | None = None) -> tuple:
        """Export this models exterior to an Optical acceptable dataset

        Args:
            faces (FaceSelection | GeoSurface): The faces to export. Defaults to None

        Returns:
            tuple: _description_
        """
        if faces is None:
            tags = self.mesh.exterior_face_tags
        else:
            tags = faces.tags

        center = np.mean(self.mesh.nodes, axis=1).squeeze()
        surface = self.basis.mesh.boundary_surface(tags, center)
        field = self.interpolate(*surface.exyz)
        vertices = surface.nodes
        triangles = surface.tris
        origin = surface._origin
        E = field.E
        H = field.H
        k0 = self.k0
        return vertices, triangles, E, H, origin, k0
    
    def optycal_antenna(self, 
                        faces: FaceSelection | GeoSurface | None = None,
                        origin: tuple[float, float, float] | None = None,
                        syms: list[Literal['Ex','Ey','Ez', 'Hx','Hy','Hz']] | None = None) -> dict:
        """Export this models exterior to an Optical acceptable dataset

        Args:
            faces (FaceSelection | GeoSurface): The faces to export. Defaults to None

        Returns:
            tuple: _description_
        """
        freq = self.freq
        def function(theta: np.ndarray, phi: np.ndarray, k0: float):
            E, H = self.farfield(theta, phi, faces, origin, syms)
            return E[0,:], E[1,:], E[2,:], H[0,:], H[1,:], H[2,:]
    
        return dict(freq=freq, ff_function=function)

    # def surface_integral(self, faces: FaceSelection | GeoSurface, fieldfunction: Callable) -> float | complex:
    #     """Computes a surface integral on the selected faces. 
    
    #     The fieldfunction argument must be a callable of a single argument x, which will
    #     be of type EHField which is restuned by the field.interpolate(x,y,z) function. It has
    #     fields like Ez, Ey, Sx etc that can be called. 

    #     Args:
    #         faces (FaceSelection | GeoSurface): _description_
    #         fieldfunction (Callable): _description_

    #     Returns:
    #         float | complex: _description_
    #     """
    #     from ...mth.integrals import surface_integral
        
    #     def ff(x, y, z):
    #         fieldobj = self.interpolate(x,y,z)
    #         return fieldfunction(fieldobj)
        
    #     nodes = self.mesh.get_nodes(faces.tags)
    #     triangles = self.mesh.get_triangles(faces.tags)
        
    #     return surface_integral(nodes, triangles, ff)
        
        
class MWScalar:
    """The MWDataSet class stores solution data of FEM Time Harmonic simulations.
    """
    _fields: list[str] = ['freq','k0','Sp','beta','Pout','Z0']
    _copy: list[str] = ['_portmap','_portnumbers','port_modes']

    def __init__(self):
        self.freq: float = None
        self.k0: float = None
        self.Sp: np.ndarray = None
        self.beta: np.ndarray = None
        self.Z0: np.ndarray = None
        self.Pout: np.ndarray = None
        self._portmap: dict[int, float|int] = dict()
        self._portnumbers: list[int | float] = []
        self.port_modes: list[PortProperties] = []

    def init_sp(self, portnumbers: list[int | float]) -> None:
        """Initialize the S-parameter dataset with the given number of ports."""
        self._portnumbers = portnumbers
        i = 0
        for n in portnumbers:
            self._portmap[n] = i
            i += 1

        self.Sp = np.zeros((i,i), dtype=np.complex128)
        self.Z0 = np.zeros((i,), dtype=np.complex128)
        self.Pout = np.zeros((i,), dtype=np.float64)
        self.beta = np.zeros((i,), dtype=np.complex128)

        
    def write_S(self, i1: int | float, i2: int | float, value: complex) -> None:
        self.Sp[self._portmap[i1], self._portmap[i2]] = value

    def S(self, i1: int, i2: int) -> complex:
        """Return the S-parameter corresponding to the given set of indices:

        S11 = obj.S(1,1)

        Args:
            i1 (int): The first port index
            i2 (int): The second port index

        Returns:
            complex: The S-parameter
        """
        return self.Sp[self._portmap[i1], self._portmap[i2]]
    
    def add_port_properties(self, 
                            port_number: int,
                            mode_number: int,
                            k0: float,
                            beta: float,
                            Z0: float | complex,
                            Pout: float) -> None:
        i = self._portmap[port_number]
        self.beta[i] = beta
        self.Z0[i] = Z0
        self.Pout[i] = Pout
    
class MWScalarNdim:
    _fields: list[str] = ['freq','k0','Sp','beta','Pout','Z0']
    _copy: list[str] = ['_portmap','_portnumbers']

    def __init__(self):
        self.freq: np.ndarray = None
        self.k0: np.ndarray = None
        self.Sp: np.ndarray = None
        self.beta: np.ndarray = None
        self.Z0: np.ndarray = None
        self.Pout: np.ndarray = None
        self._portmap: dict[int, float|int] = dict()
        self._portnumbers: list[int | float] = []

    def S(self, i1: int, i2: int) -> np.ndarray:
        return self.Sp[...,self._portmap[i1], self._portmap[i2]]
    
    @property
    def Smat(self) -> np.ndarray:
        """Returns the full S-matrix

        Returns:
            np.ndarray: The S-matrix with shape (nF, nP, nP)
        """
        Nports = len(self._portmap)
        nfreq = self.freq.shape[0]

        Smat = np.zeros((nfreq,Nports,Nports), dtype=np.complex128)
        
        for i in self._portnumbers:
            for j in self._portnumbers:
                Smat[:,i-1,j-1] = self.S(i,j)

        return Smat
    
    def emmodel(self, f_sample: np.ndarray | None = None) -> tuple[np.ndarray, np.ndarray]:
        """Returns the required date for a Heavi S-parameter component

        Returns:
            tuple[np.ndarray, np.ndarray]: Heavi data
        """
        
        if f_sample is not None:
            f = f_sample
            S = self.model_Smat(f_sample)
        else:
            f = self.freq
            S = self.Smat
        
        Z0s = self.Z0
        S = renormalise_s(S, Z0s, 50.0)
        return f, S

    def model_S(self, i: int, j: int, 
            freq: np.ndarray, 
            Npoles: int | Literal['auto'] = 'auto', 
            inc_real: bool = False,
            maxpoles: int = 30) -> np.ndarray:
        """Returns an S-parameter model object at a dense frequency range.
        This method uses vector fitting inside the datasets frequency points to determine a model for the linear system.

        Args:
            i (int): The first S-parameter index
            j (int): The second S-parameter index
            freq (np.ndarray): The frequency sample points
            Npoles (int | 'auto', optional): The number of poles to use (approx 2x divice order). Defaults to 10.
            inc_real (bool, optional): Wether to allow for a real-pole. Defaults to False.

        Returns:
            SparamModel: The SparamModel object
        """
        return SparamModel(self.freq, self.S(i,j), n_poles=Npoles, inc_real=inc_real, maxpoles=maxpoles)(freq)

    def model_Smat(self, frequencies: np.ndarray,
                        Npoles: int = 10,
                        inc_real: bool = False) -> np.ndarray:
        """Generates a full S-parameter matrix on the provided frequency points using the Vector Fitting algorithm.

        This function output can be used directly with the .save_matrix() method.

        Args:
            frequencies (np.ndarray): The sample frequencies
            Npoles (int, optional): The number of poles to fit. Defaults to 10.
            inc_real (bool, optional): Wether allow for a real pole. Defaults to False.

        Returns:
            np.ndarray: The (Nf,Np,Np) S-parameter matrix
        """
        Nports = len(self._portmap)
        nfreq = frequencies.shape[0]

        Smat = np.zeros((nfreq,Nports,Nports), dtype=np.complex128)
        
        for i in self._portnumbers:
            for j in self._portnumbers:
                S = self.model_S(i,j,frequencies, Npoles=Npoles, inc_real=inc_real)
                Smat[:,i-1,j-1] = S

        return Smat

    def export_touchstone(self, 
                            filename: str,
                            Z0ref: float | None = None,
                            format: Literal['RI','MA','DB'] = 'RI',
                            custom_comments: list[str] | None = None,
                            funit: Literal['HZ','KHZ','MHZ','GHZ'] = 'GHZ'):
        """Export the S-parameter data to a touchstone file

        This function assumes that all ports are numbered in sequence 1,2,3,4... etc with
        no missing ports. Otherwise it crashes. Will be update/improved soon with more features.

        Additionally, one may provide a reference impedance. If this argument is provided, a port impedance renormalization
        will be performed to that common impedance.

        Args:
            filename (str): The File name
            Z0ref (float): The reference impedance to normalize to. Defaults to None
            format (Literal[DB, RI, MA]): The dataformat used in the touchstone file.
            custom_comments : list[str], optional. List of custom comment strings to add to the touchstone file header.
                                                    Each string will be prefixed with "! " automatically.
        """
        
        logger.info(f'Exporting S-data to {filename}')
        Nports = len(self._portmap)
        freqs = self.freq

        Smat = np.zeros((len(freqs),Nports,Nports), dtype=np.complex128)
        
        for i in range(1,Nports+1):
            for j in range(1,Nports+1):
                S = self.S(i,j)
                Smat[:,i-1,j-1] = S
        
        self.save_smatrix(filename, Smat, freqs, format=format, Z0ref=Z0ref, custom_comments=custom_comments, funit=funit)

    def save_smatrix(self, 
                        filename: str,
                        Smatrix: np.ndarray,
                        frequencies: np.ndarray, 
                        Z0ref: float | None = None,
                        format: Literal['RI','MA','DB'] = 'RI',
                        custom_comments: list[str] | None = None,
                        funit: Literal['HZ','KHZ','MHZ','GHZ'] = 'GHZ') -> None:
        """Save an S-parameter matrix to a touchstone file.
        
        Additionally, a reference impedance may be supplied. In this case, a port renormalization will be performed on the S-matrix.

        Args:
            filename (str): The filename
            Smatrix (np.ndarray): The S-parameter matrix with shape (Nfreq, Nport, Nport)
            frequencies (np.ndarray): The frequencies with size (Nfreq,)
            Z0ref (float, optional): An optional reference impedance to normalize to. Defaults to None.
            format (Literal["RI","MA",'DB], optional): The S-parameter format. Defaults to 'RI'.
            custom_comments : list[str], optional. List of custom comment strings to add to the touchstone file header.
                                                    Each string will be prefixed with "! " automatically.
        """
        from .touchstone import generate_touchstone

        if Z0ref is not None:
            Z0s = self.Z0
            logger.debug(f'Renormalizing impedances {Z0s}Ω to {Z0ref}Ω')
            Smatrix = renormalise_s(Smatrix, Z0s, Z0ref)


        generate_touchstone(filename, frequencies, Smatrix, format, custom_comments, funit)
        
        logger.info('Export complete!')
        
# class MWSimData(SimData[MWConstants]):
#     """The MWSimData class contains all EM simulation data from a Time Harmonic simulation
#     along all sweep axes.
#     """
#     datatype: type = MWConstants
#     def __init__(self):
#         super().__init__()
#         self._injections = dict()
#         self._axis = 'freq'

#     def __getitem__(self, field: EMField) -> np.ndarray:
#         return getattr(self, field)

#     @property
#     def mesh(self) -> Mesh3D:
#         """Returns the relevant mesh object for this dataset assuming they are all the same.

#         Returns:
#             Mesh3D: The mesh object.
#         """
#         return self.datasets[0].basis.mesh
    
#     def howto(self) -> None:
#         """To access data in the MWSimData class use the .ax method to extract properties selected
#         along an access of global variables. The axes are all global properties that the MWDataSets manage.
        
#         For example the following would return all S(2,1) parameters along the frequency axis.
        
#         >>> freq, S21 = dataset.ax('freq').S(2,1)

#         Alternatively, one can manually select any solution indexed in order of generation using.

#         >>> S21 = dataset.item(3).S(2,1)

#         To find the E or H fields at any coordinate, one can use the Dataset's .interpolate method. 
#         This method returns the same Dataset object after which the computed fields can be accessed.

#         >>> Ex = dataset.item(3).interpolate(xs,ys,zs).Ex

#         Lastly, to find the solutions for a given frequency or other value, you can also just call the dataset
#         class:
        
#         >>> Ex, Ey, Ez = dataset(freq=1e9).interpolate(xs,ys,zs).E

#         """

#     def select(self, **axes: EMField) -> MWSimData:
#         """Takes the provided axis points and constructs a new dataset only for those axes values

#         Returns:
#             MWSimData: The new dataset
#         """
#         newdata = MWSimData()
#         for dataset in self.datasets:
#             if dataset.equals(**axes):
#                 newdata.datasets.append(dataset)
#         return newdata
    
#     def ax(self, *field: EMField) -> MWConstants:
#         """Return a MWDataSet proxy object that you can request properties for along a provided axis.

#         The MWSimData class contains a list of MWDataSet objects. Any global variable like .freq of the 
#         MWDataSet object can be used as inner-axes after which the outer axis can be selected as if
#         you are extract a single one.

#         Args:
#             field (EMField): The global field variable to select the data along

#         Returns:
#             MWDataSet: An MWDataSet object (actually a proxy for)

#         Example:
#         The following will select all S11 parameters along the frequency axis:

#         >>> freq, S11 = dataset.ax('freq').S(1,1)

#         """
#         # find the real DataSet
#         return _DataSetProxy(field, self.datasets)




#!/usr/bin/env python
"""
Automated installer for SuiteSparse/UMFPACK + scikit-umfpack on Windows.

Key safety features
-------------------
1.   **Explicit confirmation** for every action that modifies the system.
2.   **Dry-run preview**: run with `--dry-run` to see what *would* happen.
3.   Detects existing Conda envs and asks before overwriting.
4.   All shell commands are run with `conda run -n <env> …` to avoid
     brittle `conda activate` tricks and to isolate effects.
5.   Catches and reports subprocess errors without crashing silently.
"""

from __future__ import annotations

import argparse
import json
import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import List, Optional

# --------------------------------------------------------------------------- #
# Utility helpers
# --------------------------------------------------------------------------- #

CONTINUE = input('WARNING!! This script is auto generated by ChatGPT. It is UNTESTED and being verified' + 
                    'Continue at your own risk! Do you want to proceed? [y/n]')
if CONTINUE != 'y':
    print('Aborting')
    quit()

def ask_yes_no(prompt: str, default: str | None = "y") -> bool:
    """Prompt the user with a yes/no question."""
    if default not in {"y", "n", None}:
        raise ValueError("default must be 'y', 'n', or None")
    suffix = " [Y/n] " if default == "y" else " [y/N] " if default == "n" else " [y/n] "
    while True:
        ans = input(prompt + suffix).strip().lower()
        if not ans and default:
            ans = default
        if ans in {"y", "yes"}:
            return True
        if ans in {"n", "no"}:
            return False
        print("Please enter y or n.")

def run(cmd: str | List[str], *, dry: bool = False, cwd: str | Path | None = None) -> None:
    """Run a shell command with live output, honor dry-run mode."""
    print(f"\n>>> {cmd}")
    if dry:
        return
    try:
        subprocess.run(cmd, shell=isinstance(cmd, str), check=True, cwd=cwd)
    except subprocess.CalledProcessError as exc:
        print(f"❌ Command failed (exit {exc.returncode}). Aborting.")
        sys.exit(exc.returncode)

def find_conda_root() -> Optional[Path]:
    """Guess common Conda installation directories."""
    guesses = [
        Path(os.environ.get("CONDA_PREFIX", "")),                                 # active env
        Path(os.environ.get("USERPROFILE", "")) / "Miniconda3",
        Path(os.environ.get("USERPROFILE", "")) / "Anaconda3",
    ]
    for p in guesses:
        if (p / "condabin" / "conda.bat").exists():
            return p
    return None

# --------------------------------------------------------------------------- #
# Installer logic
# --------------------------------------------------------------------------- #

def write_nativefile(conda_root: Path, *, dry: bool = False) -> Path:
    """Create a nativefile.ini pointing to SuiteSparse headers/libs in Conda."""
    content = f"""
[binaries]
c = 'x86_64-w64-mingw32-gcc'
cpp = 'x86_64-w64-mingw32-g++'

[properties]
umfpack-libdir = '''{conda_root.as_posix()}/Library/lib'''
umfpack-includedir = '''{conda_root.as_posix()}/Library/include/suitesparse'''
""".lstrip()

    nativefile = Path.cwd() / "nativefile.ini"
    print(f"\n📝 Creating {nativefile} …")
    if dry:
        print(content)
    else:
        nativefile.write_text(content, encoding="utf-8")
    return nativefile

def conda_run(env: str, args: str) -> str:
    """Helper to prepend 'conda run -n <env>'."""
    return f"conda run -n {env} {args}"

def main() -> None:
    parser = argparse.ArgumentParser(description="Safe UMFPACK installer")
    parser.add_argument(
        "--env",
        default="umf-env",
        help="Name of the Conda environment to create/use (default: %(default)s)",
    )
    parser.add_argument("--python", default="3.10", help="Python version for the env")
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Show commands but do not execute them",
    )
    args = parser.parse_args()
    dry = args.dry_run
    env_name = args.env
    python_ver = args.python

    print(
        "🛠️  UMFPACK/scikit-umfpack automated installer\n"
        "--------------------------------------------------------"
    )

    # 1. Locate Conda
    conda_root = find_conda_root()
    if conda_root:
        print(f"✅ Found Conda at {conda_root}")
    else:
        print("⚠️  Could not auto-detect Conda. Make sure it is on PATH.")
        conda_path_str = input("Paste the full path to your Conda root (e.g. C:/Miniconda3): ").strip()
        conda_root = Path(conda_path_str)
        if not (conda_root / "condabin" / "conda.bat").exists():
            print("❌ The provided path does not look like a Conda installation. Aborting.")
            sys.exit(1)

    # 2. Create or reuse environment
    print(f"\n🔍 Checking if environment '{env_name}' exists …")
    envs_json = subprocess.check_output(["conda", "env", "list", "--json"], text=True)
    envs = json.loads(envs_json)["envs"]
    env_path = next((Path(p) for p in envs if Path(p).name == env_name), None)

    if env_path:
        print(f"ℹ️  Conda env '{env_name}' already exists at {env_path}.")
        if not ask_yes_no("Reuse this environment? Otherwise I can recreate it", default="y"):
            if ask_yes_no(f"Delete existing env '{env_name}' and recreate it?", default="n"):
                run(f"conda remove -n {env_name} --all -y", dry=dry)
            else:
                print("Aborting as per user request.")
                return

    if not env_path:
        if ask_yes_no(f"Create new environment '{env_name}' with Python {python_ver}?", default="y"):
            run(f"conda create -n {env_name} python={python_ver} -y", dry=dry)
        else:
            print("Aborting at user request.")
            return

    # 3. Install core dependencies
    print("\n📦 Installing Conda packages (this may take a while) …")
    conda_pkgs = [
        "anaconda::suitesparse",
        "meson",
        "swig",
        "-c conda-forge compilers",
        "-c conda-forge m2w64-toolchain",
        "-c conda-forge openblas",
    ]
    run(conda_run(env_name, f"conda install {' '.join(conda_pkgs)} -y"), dry=dry)

    # 4. Install pip dependencies
    print("\n📦 Installing pip packages …")
    pip_cmds = [
        "python -m pip install --upgrade pip",
        "python -m pip install meson-python",
        "python -m pip install numpy scipy",
    ]
    for cmd in pip_cmds:
        run(conda_run(env_name, cmd), dry=dry)

    # 5. Write nativefile.ini
    nativefile_path = write_nativefile(conda_root, dry=dry)

    # 6. Build & install scikit-umfpack
    print("\n🔧 Building scikit-umfpack from source …")
    build_cmd = (
        "python -m pip install scikit-umfpack "
        "--no-build-isolation "
        f"-Csetup-args=\"--native-file={nativefile_path}\""
    )
    run(conda_run(env_name, build_cmd), dry=dry)

    # 7. Optional extras
    extras = [
        "python -m pip install numba numba-progress --no-binary :all:",
        "python -m pip install emerge --no-deps",
    ]
    if ask_yes_no("Install optional extras (numba, numba-progress, emerge)?", default="y"):
        for cmd in extras:
            run(conda_run(env_name, cmd), dry=dry)

    print(
        "\n🎉 Installation finished.\n"
        "Test it with:\n"
        f"   conda run -n {env_name} python - <<'PY'\n"
        "import scikits.umfpack, sys; print('UMFPACK OK on', sys.platform)\n"
        "PY"
    )

# --------------------------------------------------------------------------- #

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2.0-or-later
"""
adaptive_frequency.py
---------------------
Copyright (c) 2014  Phil Reinhold
Copyright (c) 2025  Robert Fennis

This file is part of EMerge.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License **version 2**,
or (at your option) **any later version**, as published by the Free
Software Foundation.

This program is distributed in the hope that it will be useful,
but **WITHOUT ANY WARRANTY**; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (see the file “COPYING” in the project root).
If not, see <https://www.gnu.org/licenses/>.

-------------------------------------------------------------------------------
Origins
-------------------------------------------------------------------------------
This module is based on “Duplication of the vector fitting algorithm in
Python” (GPL-2.0) by Phil Reinhold, which re-implements the original MATLAB
code by **Bjørn Gustavsen**.  Key academic references:

  [1] B. Gustavsen & A. Semlyen, “Rational approximation of frequency
      domain responses by Vector Fitting”, *IEEE Trans. Power Delivery*,
      14 (3): 1052-1061, Jul 1999.

  [2] B. Gustavsen, “Improving the pole relocating properties of vector
      fitting”, *IEEE Trans. Power Delivery*, 21 (3): 1587-1592, Jul 2006.

  [3] D. Deschrijver *et al.*, “Macromodeling of Multiport Systems Using a
      Fast Implementation of the Vector Fitting Method”, *IEEE MWC Lett.*,
      18 (6): 383-385, Jun 2008.

-------------------------------------------------------------------------------
Modification history
-------------------------------------------------------------------------------
* 2025-06-23  Robert Fennis  fennisrobert@gmail.com
      - Refactored API with logic realized in the SparamModel class.

"""

# ty: ignore

import numpy as np
from typing import Literal
from loguru import logger

def cc(z):
    return z.conjugate()

def model(s, poles, residues, d, h):
    return sum([r/(s-p) for p, r in zip(poles, residues)]) + d + s*h

def vectfit_step(f: np.ndarray, s: np.ndarray, poles: np.ndarray) -> np.ndarray:
    """
    f = complex data to fit
    s = j*frequency
    poles = initial poles guess
        note: All complex poles must come in sequential complex conjugate pairs
    returns adjusted poles
    """
    N = len(poles)
    Ns = len(s)

    cindex = np.zeros(N)
    # cindex is:
    #   - 0 for real poles
    #   - 1 for the first of a complex-conjugate pair
    #   - 2 for the second of a cc pair
    for i, p in enumerate(poles):
        if p.imag != 0:
            if i == 0 or cindex[i-1] != 1:
                assert cc(poles[i]) == poles[i+1], ("Complex poles must come in conjugate pairs: %s, %s" % (poles[i], poles[i+1]))
                cindex[i] = 1
            else:
                cindex[i] = 2

    # First linear equation to solve. See Appendix A
    A = np.zeros((Ns, 2*N+2), dtype=np.complex128)
    for i, p in enumerate(poles):
        if cindex[i] == 0:
            A[:, i] = 1/(s - p)
        elif cindex[i] == 1:
            A[:, i] = 1/(s - p) + 1/(s - cc(p))
        elif cindex[i] == 2:
            A[:, i] = 1j/(s - p) - 1j/(s - cc(p))
        else:
            raise ValueError("cindex[%s] = %s" % (i, cindex[i]))

        A [:, N+2+i] = -A[:, i] * f

    A[:, N] = 1
    A[:, N+1] = s

    # Solve Ax == b using pseudo-inverse
    b = f
    A = np.vstack((A.real, A.imag))
    b = np.concatenate((b.real, b.imag))
    x, residuals, rnk, s = np.linalg.lstsq(A, b, rcond=-1)

    # We only want the "tilde" part in (A.4)
    x = x[-N:]

    # Calculation of zeros: Appendix B
    A = np.diag(poles)
    b = np.ones(N)
    c = x
    for i, (ci, p) in enumerate(zip(cindex, poles)):
        if ci == 1:
            x, y = p.real, p.imag
            A[i, i] = A[i+1, i+1] = x
            A[i, i+1] = -y
            A[i+1, i] = y
            b[i] = 2
            b[i+1] = 0

    H = A - np.outer(b, c)
    H = H.real
    new_poles = np.sort(np.linalg.eigvals(H))
    unstable = np.real(new_poles) > 0
    new_poles[unstable] -= 2*np.real(new_poles)[unstable]
    return new_poles

# Dear gods of coding style, I sincerely apologize for the following copy/paste
def calculate_residues(f: np.ndarray, s: np.ndarray, poles: np.ndarray, rcond=-1) -> tuple[np.ndarray, float, float]:
    Ns = len(s)
    N = len(poles)

    cindex = np.zeros(N)
    for i, p in enumerate(poles):
        if p.imag != 0:
            if i == 0 or cindex[i-1] != 1:
                assert cc(poles[i]) == poles[i+1], ("Complex poles must come in conjugate pairs: %s, %s" % poles[i:i+1])
                cindex[i] = 1
            else:
                cindex[i] = 2

    # use the new poles to extract the residues
    A = np.zeros((Ns, N+2), dtype=np.complex128)
    for i, p in enumerate(poles):
        if cindex[i] == 0:
            A[:, i] = 1/(s - p)
        elif cindex[i] == 1:
            A[:, i] = 1/(s - p) + 1/(s - cc(p))
        elif cindex[i] == 2:
            A[:, i] = 1j/(s - p) - 1j/(s - cc(p))
        else:
            raise RuntimeError("cindex[%s] = %s" % (i, cindex[i]))

    A[:, N] = 1
    A[:, N+1] = s
    # Solve Ax == b using pseudo-inverse
    b = f
    A = np.vstack((A.real, A.imag))
    b = np.concatenate((b.real, b.imag))
    cA = np.linalg.cond(A)
    if cA > 1e13:
        logger.warning('Warning!: Ill Conditioned Matrix. Consider scaling the problem down')
        logger.warning('Cond(A)', cA)
    x, residuals, rnk, s = np.linalg.lstsq(A, b, rcond=rcond)

    # Recover complex values
    x = x.astype(np.complex128)
    for i, ci in enumerate(cindex):
       if ci == 1:
           r1, r2 = x[i:i+2]
           x[i] = r1 - 1j*r2
           x[i+1] = r1 + 1j*r2

    residues = x[:N]
    d = x[N].real
    h = x[N+1].real
    return residues, d, h

def vectfit_auto(f: np.ndarray, 
                 s: np.ndarray, 
                 n_poles: int = 10, 
                 n_iter: int = 10,
                 inc_real: bool = False, 
                 loss_ratio: float = 1e-2, 
                 rcond: int = -1, 
                 track_poles: bool = False) -> tuple[np.ndarray, np.ndarray, float, float, np.ndarray]:
    w = s.imag
    pole_locs = np.linspace(w[0], w[-1], n_poles+2)[1:-1]
    lr = loss_ratio
    poles = np.concatenate([[p*(-lr + 1j), p*(-lr - 1j)] for p in pole_locs])

    if inc_real:
        poles = np.concatenate((poles, [1]))

    poles_list = []
    for _ in range(n_iter):
        poles = vectfit_step(f, s, poles)
        poles_list.append(poles)

    residues, d, h = calculate_residues(f, s, poles, rcond=rcond)

    if track_poles:
        return poles, residues, d, h, np.array(poles_list)
    return poles, residues, d, h, np.array([])

def vectfit_auto_rescale(f: np.ndarray, s: np.ndarray, 
                         n_poles: int = 10, 
                         n_iter: int = 10,
                         inc_real: bool = False, 
                         loss_ratio: float = 1e-2, 
                         rcond: int = -1, 
                         track_poles: bool = False) -> tuple[np.ndarray, np.ndarray, float, float, np.ndarray]:
    s_scale = abs(s[-1])
    f_scale = abs(f[-1])
    poles_s, residues_s, d_s, h_s, tracked_poles = vectfit_auto(f / f_scale, 
                                                 s / s_scale,
                                                  n_poles=n_poles,
                                                  n_iter = n_iter,
                                                  inc_real=inc_real,
                                                  loss_ratio=loss_ratio,
                                                  rcond=rcond,
                                                  track_poles=track_poles)
    poles = poles_s * s_scale
    residues = residues_s * f_scale * s_scale
    d = d_s * f_scale
    h = h_s * f_scale / s_scale
    return poles, residues, d, h, tracked_poles


class SparamModel:

    def __init__(self, 
                 frequencies: np.ndarray,
                 Sparam: np.ndarray,
                 n_poles: int | Literal['auto'] = 10,
                 inc_real: bool = False,
                 maxpoles: int = 40):
        self.f: np.ndarray = frequencies
        self.S: np.ndarray = Sparam

        s = 1j*frequencies

        if n_poles == 'auto':
            fdense = np.linspace(min(self.f), max(self.f), max(201, 10*self.f.shape[0]))
            success = False
            for nps in range(1,maxpoles):
                poles, residues, d, h, _ = vectfit_auto_rescale(Sparam, s, n_poles=nps, inc_real=inc_real)
                self.poles: np.ndarray = poles
                self.residues: np.ndarray = residues
                self.d = d
                self.h = h

                S = self(fdense)

                error = np.mean(np.abs(Sparam-self(self.f)))
                if all(np.abs(S) <= 1.0) and error < 1e-2:
                    logger.debug(f'Using {nps} poles.')
                    success = True
                    break
            if not success:
                logger.warning('Could not model S-parameters. Try a denser grid')

        else:
            poles, residues, d, h, _ = vectfit_auto_rescale(Sparam, s, n_poles=n_poles, inc_real=inc_real)
            self.poles: np.ndarray = poles
            self.residues: np.ndarray = residues
            self.d = d
            self.h = h
    
    def __call__(self, f: np.ndarray) -> np.ndarray:
        return model(1j*f, self.poles, self.residues, self.d, self.h)
    


from ..mesh3d import Mesh3D
from ..geometry import GeoObject
from ..selection import Selection
from ..physics.microwave.microwave_bc import PortBC
from typing import Iterable, Literal
import numpy as np

cmap_names = Literal['bgy','bgyw','kbc','blues','bmw','bmy','kgy','gray','dimgray','fire','kb','kg','kr',
                     'bkr','bky','coolwarm','gwv','bjy','bwy','cwr','colorwheel','isolum','rainbow','fire',
                     'cet_fire','gouldian','kbgyw','cwr','CET_CBL1','CET_CBL3','CET_D1A']


class BaseDisplay:

    def __init__(self, mesh: Mesh3D):
        self._mesh: Mesh3D = mesh

    def show(self):
        raise NotImplementedError('This method is not implemented')
        
    def add_object(self, obj: GeoObject | Selection,*args, **kwargs):
        """ Adds an object to the display

        Keyword arguments
        ----------
        color : ColorLike, optional
            Use to make the entire mesh have a single solid color.
            Either a string, RGB list, or hex color string.  For example:
            ``color='white'``, ``color='w'``, ``color=[1.0, 1.0, 1.0]``, or
            ``color='#FFFFFF'``. Color will be overridden if scalars are
            specified.

            Defaults to :attr:`pyvista.global_theme.color
            <pyvista.plotting.themes.Theme.color>`.

        style : str, optional
            Visualization style of the mesh.  One of the following:
            ``style='surface'``, ``style='wireframe'``, ``style='points'``,
            ``style='points_gaussian'``. Defaults to ``'surface'``. Note that
            ``'wireframe'`` only shows a wireframe of the outer geometry.
            ``'points_gaussian'`` can be modified with the ``emissive``,
            ``render_points_as_spheres`` options.

        scalars : str | numpy.ndarray, optional
            Scalars used to "color" the mesh.  Accepts a string name
            of an array that is present on the mesh or an array equal
            to the number of cells or the number of points in the
            mesh.  Array should be sized as a single vector. If both
            ``color`` and ``scalars`` are ``None``, then the active
            scalars are used.

        clim : sequence[float], optional
            Two item color bar range for scalars.  Defaults to minimum and
            maximum of scalars array.  Example: ``[-1, 2]``. ``rng`` is
            also an accepted alias for this.

        show_edges : bool, optional
            Shows the edges of a mesh.  Does not apply to a wireframe
            representation.

        edge_color : ColorLike, optional
            The solid color to give the edges when ``show_edges=True``.
            Either a string, RGB list, or hex color string.

            Defaults to :attr:`pyvista.global_theme.edge_color
            <pyvista.plotting.themes.Theme.edge_color>`.

        point_size : float, optional
            Point size of any nodes in the dataset plotted. Also
            applicable when style='points'. Default ``5.0``.

        line_width : float, optional
            Thickness of lines.  Only valid for wireframe and surface
            representations.  Default ``None``.

        opacity : float | str| array_like
            Opacity of the mesh. If a single float value is given, it
            will be the global opacity of the mesh and uniformly
            applied everywhere - should be between 0 and 1. A string
            can also be specified to map the scalars range to a
            predefined opacity transfer function (options include:
            ``'linear'``, ``'linear_r'``, ``'geom'``, ``'geom_r'``).
            A string could also be used to map a scalars array from
            the mesh to the opacity (must have same number of elements
            as the ``scalars`` argument). Or you can pass a custom
            made transfer function that is an array either
            ``n_colors`` in length or shorter.

        flip_scalars : bool, default: False
            Flip direction of cmap. Most colormaps allow ``*_r``
            suffix to do this as well.

        lighting : bool, optional
            Enable or disable view direction lighting. Default ``False``.

        n_colors : int, optional
            Number of colors to use when displaying scalars. Defaults to 256.
            The scalar bar will also have this many colors.

        interpolate_before_map : bool, optional
            Enabling makes for a smoother scalars display.  Default is
            ``True``.  When ``False``, OpenGL will interpolate the
            mapped colors which can result is showing colors that are
            not present in the color map.

        cmap : str | list | pyvista.LookupTable, default: :attr:`pyvista.plotting.themes.Theme.cmap`
            If a string, this is the name of the ``matplotlib`` colormap to use
            when mapping the ``scalars``.  See available Matplotlib colormaps.
            Only applicable for when displaying ``scalars``.
            ``colormap`` is also an accepted alias
            for this. If ``colorcet`` or ``cmocean`` are installed, their
            colormaps can be specified by name.

            You can also specify a list of colors to override an existing
            colormap with a custom one.  For example, to create a three color
            colormap you might specify ``['green', 'red', 'blue']``.

            This parameter also accepts a :class:`pyvista.LookupTable`. If this
            is set, all parameters controlling the color map like ``n_colors``
            will be ignored.

        label : str, optional
            String label to use when adding a legend to the scene with
            :func:`pyvista.Plotter.add_legend`.

        reset_camera : bool, optional
            Reset the camera after adding this mesh to the scene. The default
            setting is ``None``, where the camera is only reset if this plotter
            has already been shown. If ``False``, the camera is not reset
            regardless of the state of the ``Plotter``. When ``True``, the
            camera is always reset.

        scalar_bar_args : dict, optional
            Dictionary of keyword arguments to pass when adding the
            scalar bar to the scene. For options, see
            :func:`pyvista.Plotter.add_scalar_bar`.

        show_scalar_bar : bool, optional
            If ``False``, a scalar bar will not be added to the
            scene.

        multi_colors : bool | str | cycler.Cycler | sequence[ColorLike], default: False
            If a :class:`pyvista.MultiBlock` dataset is given this will color
            each block by a solid color using a custom cycler.

            If ``True``, the default 'matplotlib' color cycler is used.

            See :func:`set_color_cycler<Plotter.set_color_cycler>` for usage of
            custom color cycles.

        name : str, optional
            The name for the added mesh/actor so that it can be easily
            updated.  If an actor of this name already exists in the
            rendering window, it will be replaced by the new actor.

        texture : pyvista.Texture or np.ndarray, optional
            A texture to apply if the input mesh has texture
            coordinates.  This will not work with MultiBlock
            datasets.

        render_points_as_spheres : bool, optional
            Render points as spheres rather than dots.

        render_lines_as_tubes : bool, optional
            Show lines as thick tubes rather than flat lines.  Control
            the width with ``line_width``.

        smooth_shading : bool, optional
            Enable smooth shading when ``True`` using the Phong
            shading algorithm.  When ``False``, use flat shading.
            Automatically enabled when ``pbr=True``.  See
            :ref:`shading_example`.

        split_sharp_edges : bool, optional
            Split sharp edges exceeding 30 degrees when plotting with smooth
            shading.  Control the angle with the optional keyword argument
            ``feature_angle``.  By default this is ``False`` unless overridden
            by the global or plotter theme.  Note that enabling this will
            create a copy of the input mesh within the plotter.  See
            :ref:`shading_example`.

        ambient : float, optional
            When lighting is enabled, this is the amount of light in
            the range of 0 to 1 (default 0.0) that reaches the actor
            when not directed at the light source emitted from the
            viewer.

        diffuse : float, optional
            The diffuse lighting coefficient. Default 1.0.

        specular : float, optional
            The specular lighting coefficient. Default 0.0.

        specular_power : float, optional
            The specular power. Between 0.0 and 128.0.

        nan_color : ColorLike, optional
            The color to use for all ``NaN`` values in the plotted
            scalar array.

        nan_opacity : float, optional
            Opacity of ``NaN`` values.  Should be between 0 and 1.
            Default 1.0.

        culling : str, optional
            Does not render faces that are culled. Options are
            ``'front'`` or ``'back'``. This can be helpful for dense
            surface meshes, especially when edges are visible, but can
            cause flat meshes to be partially displayed.  Defaults to
            ``False``.

        rgb : bool, optional
            If an 2 dimensional array is passed as the scalars, plot
            those values as RGB(A) colors. ``rgba`` is also an
            accepted alias for this.  Opacity (the A) is optional.  If
            a scalars array ending with ``"_rgba"`` is passed, the default
            becomes ``True``.  This can be overridden by setting this
            parameter to ``False``.

        categories : bool, optional
            If set to ``True``, then the number of unique values in
            the scalar array will be used as the ``n_colors``
            argument.

        silhouette : dict, bool, optional
            If set to ``True``, plot a silhouette highlight for the
            mesh. This feature is only available for a triangulated
            ``PolyData``.  As a ``dict``, it contains the properties
            of the silhouette to display:

                * ``color``: ``ColorLike``, color of the silhouette
                * ``line_width``: ``float``, edge width
                * ``opacity``: ``float`` between 0 and 1, edge transparency
                * ``feature_angle``: If a ``float``, display sharp edges
                  exceeding that angle in degrees.
                * ``decimate``: ``float`` between 0 and 1, level of decimation

        use_transparency : bool, optional
            Invert the opacity mappings and make the values correspond
            to transparency.

        below_color : ColorLike, optional
            Solid color for values below the scalars range
            (``clim``). This will automatically set the scalar bar
            ``below_label`` to ``'below'``.

        above_color : ColorLike, optional
            Solid color for values below the scalars range
            (``clim``). This will automatically set the scalar bar
            ``above_label`` to ``'above'``.

        annotations : dict, optional
            Pass a dictionary of annotations. Keys are the float
            values in the scalars range to annotate on the scalar bar
            and the values are the string annotations.

        pickable : bool, optional
            Set whether this actor is pickable.

        preference : str, default: "point"
            When ``mesh.n_points == mesh.n_cells`` and setting
            scalars, this parameter sets how the scalars will be
            mapped to the mesh.  Default ``'point'``, causes the
            scalars will be associated with the mesh points.  Can be
            either ``'point'`` or ``'cell'``.

        log_scale : bool, default: False
            Use log scale when mapping data to colors. Scalars less
            than zero are mapped to the smallest representable
            positive float.

        pbr : bool, optional
            Enable physics based rendering (PBR) if the mesh is
            ``PolyData``.  Use the ``color`` argument to set the base
            color.

        metallic : float, optional
            Usually this value is either 0 or 1 for a real material
            but any value in between is valid. This parameter is only
            used by PBR interpolation.

        roughness : float, optional
            This value has to be between 0 (glossy) and 1 (rough). A
            glossy material has reflections and a high specular
            part. This parameter is only used by PBR
            interpolation.

        render : bool, default: True
            Force a render when ``True``.

        user_matrix : np.ndarray | vtk.vtkMatrix4x4, default: np.eye(4)
            Matrix passed to the Actor class before rendering. This affects the
            actor/rendering only, not the input volume itself. The user matrix is the
            last transformation applied to the actor before rendering. Defaults to the
            identity matrix.

        component : int, optional
            Set component of vector valued scalars to plot.  Must be
            nonnegative, if supplied. If ``None``, the magnitude of
            the vector is plotted.

        emissive : bool, optional
            Treat the points/splats as emissive light sources. Only valid for
            ``style='points_gaussian'`` representation.

        copy_mesh : bool, default: False
            If ``True``, a copy of the mesh will be made before adding it to
            the plotter.  This is useful if you would like to add the same
            mesh to a plotter multiple times and display different
            scalars. Setting ``copy_mesh`` to ``False`` is necessary if you
            would like to update the mesh after adding it to the plotter and
            have these updates rendered, e.g. by changing the active scalars or
            through an interactive widget. This should only be set to ``True``
            with caution. Defaults to ``False``. This is ignored if the input
            is a ``vtkAlgorithm`` subclass.

        backface_params : dict | pyvista.Property, optional
            A :class:`pyvista.Property` or a dict of parameters to use for
            backface rendering. This is useful for instance when the inside of
            oriented surfaces has a different color than the outside. When a
            :class:`pyvista.Property`, this is directly used for backface
            rendering. When a dict, valid keys are :class:`pyvista.Property`
            attributes, and values are corresponding values to use for the
            given property. Omitted keys (or the default of
            ``backface_params=None``) default to the corresponding frontface
            properties.

        show_vertices : bool, optional
            When ``style`` is not ``'points'``, render the external surface
            vertices. The following optional keyword arguments may be used to
            control the style of the vertices:

            * ``vertex_color`` - The color of the vertices
            * ``vertex_style`` - Change style to ``'points_gaussian'``
            * ``vertex_opacity`` - Control the opacity of the vertices

        edge_opacity : float, optional
            Edge opacity of the mesh. A single float value that will be applied globally
            edge opacity of the mesh and uniformly applied everywhere - should be
            between 0 and 1.

            .. note::
                `edge_opacity` uses ``SetEdgeOpacity`` as the underlying method which
                requires VTK version 9.3 or higher. If ``SetEdgeOpacity`` is not
                available, `edge_opacity` is set to 1.

        **kwargs : dict, optional
            Optional keyword arguments.
        """
        raise NotImplementedError('This method is not implemented')

    def add_scatter(self, xs: np.ndarray, ys: np.ndarray, zs: np.ndarray):
        raise NotImplementedError('This method is not implemented')

    def add_portmode(self, port: PortBC, 
                     Npoints: int = 10, 
                     dv=(0,0,0), 
                     XYZ=None,
                     field: Literal['E','H'] = 'E', 
                     k0: float | None = None,
                     mode_number: int | None = None):
        raise NotImplementedError('This method is not implemented')

    def add_quiver(self, x: np.ndarray, y: np.ndarray, z: np.ndarray,
              dx: np.ndarray, dy: np.ndarray, dz: np.ndarray,
              scale: float = 1,
              color: tuple[float, float, float] | None = None,
              scalemode: Literal['lin','log'] = 'lin') -> None:
        
        raise NotImplementedError('This method is not implemented')
    
    def add_surf(self, 
                 x: np.ndarray,
                 y: np.ndarray,
                 z: np.ndarray,
                 field: np.ndarray,
                 scale: Literal['lin','log','symlog'] = 'lin',
                 cmap: cmap_names = 'coolwarm',
                 clim: tuple[float, float] | None = None,
                 opacity: float = 1.0,
                 symmetrize: bool = True,
                 _fieldname: str | None = None,
                 **kwargs,):
        """Add a surface plot to the display
        The X,Y,Z coordinates must be a 2D grid of data points. The field must be a real field with the same size.

        Args:
            x (np.ndarray): The X-grid array
            y (np.ndarray): The Y-grid array
            z (np.ndarray): The Z-grid array
            field (np.ndarray): The scalar field to display
            scale (Literal["lin","log","symlog"], optional): The colormap scaling¹. Defaults to 'lin'.
            cmap (cmap_names, optional): The colormap. Defaults to 'coolwarm'.
            clim (tuple[float, float], optional): Specific color limits (min, max). Defaults to None.
            opacity (float, optional): The opacity of the surface. Defaults to 1.0.
            symmetrize (bool, optional): Wether to force a symmetrical color limit (-A,A). Defaults to True.
        
        (¹): lin: f(x)=x, log: f(x)=log₁₀(|x|), symlog: f(x)=sgn(x)·log₁₀(1+|x·ln(10)|)
        """
        raise NotImplementedError('This method is not implemented')